/* Statistics stuff */

#include <cstdlib>
#include <ctime>
#include "basedefs.h"
#include "BaseList.h"
#include "cmatrix_utils.h"
#include <iostream>
#include "lcm_basethreads.h"

namespace libcmatrix {

static int seedval=-1;

// Assume that seedval can be set atomically
void set_seed(int val)
{
  seedval=-std::abs(val);
}

void set_seed()
{
  seedval=-int(time(NULL) &0x7fff);
}

/* Routines taken from Numerical Recipes - new interface to gauss() */

#define M1 259200
#define IA1 7141
#define IC1 54773
#define RM1 (1.0/M1)
#define M2 134456
#define IA2 8121
#define IC2 28411
#define RM2 (1.0/M2)
#define M3 243000
#define IA3 4561
#define IC3 51349

static double nr_random_unsafe(int *idum)
{
  static long ix1,ix2,ix3;
  static double r[98];
  double temp;
  static bool iff=false;
  int j;
  
  if (*idum < 0 || !iff) {
    iff=true;
    ix1=(IC1-(*idum)) % M1;
    ix1=(IA1*ix1+IC1) % M1;
    ix2=ix1 % M2;
    ix1=(IA1*ix1+IC1) % M1;
    ix3=ix1 % M3;
    for (j=1;j<=97;j++) {
      ix1=(IA1*ix1+IC1) % M1;
      ix2=(IA2*ix2+IC2) % M2;
      r[j]=(ix1+ix2*RM2)*RM1;
    }
    *idum=1;
  }
  ix1=(IA1*ix1+IC1) % M1;
  ix2=(IA2*ix2+IC2) % M2;
  ix3=(IA3*ix3+IC3) % M3;
  j=1 + ((97*ix3)/M3);
  if (j > 97 || j < 1)
    throw InternalError("nr_random");
  temp=r[j];
  r[j]=(ix1+ix2*RM2)*RM1;
  return temp;
}

double random(double sigma, bool unsafe)
{
  static Mutex<ThreadingActive> lock;
  if (!unsafe)
    lock.acquire();
  const double res=sigma*nr_random_unsafe(&seedval);
  if (!unsafe)
    lock.release();
  return res;
}

double gauss(double sigma, bool unsafe)
{
  static Mutex<ThreadingActive> lock;
  if (!unsafe)
    lock.acquire();

  static bool iset=false;
  static double gset;
  double fac,r,v1,v2,res;
  
  if  (iset == false) {
    do {
      v1=2.0*nr_random_unsafe(&seedval)-1.0;
      v2=2.0*nr_random_unsafe(&seedval)-1.0;
      r=v1*v1+v2*v2;
    } while (r >= 1.0);
    fac=std::sqrt(-2.0*std::log(r)/r);
    gset=v1*fac;
    iset=true;
    res=sigma*v2*fac;
  }
  else {
    iset=false;
    res=sigma*gset;
  }
  if (!unsafe)
    lock.release();

  return res;
}

double nr_random(int *idum, bool unsafe)
{
  static Mutex<ThreadingActive> lock;

  if (!unsafe)
    lock.acquire();
  const double res=nr_random_unsafe(idum);
  if (!unsafe)
    lock.release();
  return res;
}

double stddev(const BaseList<double>& a)
{
  if (a.size()<3)
    throw Failed("stddev: called with data set of <3 values!");
  const size_t n=a.size();
  const double mean=sum(a)/n;
  double tot=0.0;
  for (size_t i=n;i--;) {
    const double diff=a(i)-mean;
    tot+=diff*diff;
  }
  return std::sqrt(tot/n); // fix Jan 07
}

}//namespace libcmatrix
