/*! \file
  \brief  Processing directives for 2D processing
*/

#include "NMRsim_Process.h"

class ProcessSplitSinCos : public ProcessCommand {
public:
  ProcessSplitSinCos()
    : ProcessCommand(PROC_HAS2D | PROC_CHANGES) {}
  void exec(cmatrix&, BaseList<processing_state>) const;
  static ProcessCommand* create() { return new ProcessSplitSinCos(); }
  void print(std::ostream& ostr) const {  ostr << "splitsincos"; }
  static Warning<> fd_warning; //!< applying split to frequency domain data
};

class ProcessFTIndirect : public ProcessCommand {
public:
  ProcessFTIndirect(size_t flagsv)
    : ProcessCommand(PROC_HAS2D), flags_(flagsv) {}
  void exec(cmatrix&, BaseList<processing_state>) const;
  static ProcessCommand* create();
  void print(std::ostream&) const;
private:
  size_t flags_;
};

class ProcessBaselineCorrect : public ProcessCommand {
public:
  ProcessBaselineCorrect(const BaseList<size_t>& indsv)
    : ProcessCommand(PROC_HAS1D), inds_(indsv) {
    if (inds_.empty())
      throw InvalidParameter("ProcessBaselineCorrect()");
  }
  static ProcessCommand* create();
  void exec(BaseList<complex>, processing_state&) const;
  void print(std::ostream& ostr) const { ostr << "baselinecorrect " << inds_; }
private:
  LIST<size_t> inds_;
};

class ProcessAdd : public ProcessCommand {
public:
  ProcessAdd(const char*, double =1.0);
  void exec(cmatrix&, BaseList<processing_state>) const;
  static ProcessCommand* create();
  void print(std::ostream&) const;
private:
  std::string name_;
  cmatrix data_;
  filestruct pars_;
  double scale_;
};

namespace {
  struct Proxy_ {
    Proxy_() {
      Process_Factory_t& factory(get_Process_Factory());

      factory["splitsincos"]=&ProcessSplitSinCos::create;
      factory["ftindirect"]=&ProcessFTIndirect::create;
      factory["add"]=&ProcessAdd::create;
      factory["baselinecorrect"]=&ProcessBaselineCorrect::create;
    }
  };
  
  Proxy_ proxy;
}

ProcessAdd::ProcessAdd(const char* namev, double scalev)
  : ProcessCommand(PROC_HAS2D),
    name_(namev), scale_(scalev)
{
  raw_read_file(data_,pars_,namev);
}

void ProcessAdd::print(std::ostream& ostr) const
{
  ostr << "add " << name_;
  if (scale_!=1.0)
    ostr << ' ' << scale_;
}

Warning<> combine_sw_warning("apparently mismatched spectral widths",&NMRsim_repeat_warning);
Warning<> combine_domain_warning("current and added data set apparently in different domains - dimension: ",&NMRsim_repeat_warning);

void checkdomainsw(const processing_state& pflag, double sw, domain_t dom, const char* domname)
{
  if (sw && pflag.sw) {
    if (fabs(sw-pflag.sw)>1e-5*sw) {
      char buf[256];
      snprintf(buf,sizeof(buf),": sw of current data set=%g kHz, sw of added data set=%g kHz (%s dimension)",pflag.sw*1e-3,sw*1e-3,domname);
      combine_sw_warning.raise(buf);
    }
  }
  if (dom && (pflag.istimedomain!=(dom==D_TIME)))
    combine_domain_warning.raise(domname);
}

void ProcessAdd::exec(cmatrix& a, BaseList<processing_state> pflags) const
{
  if (!arematching(a,data_)) {
    std::cerr << "Can't add data set " << name_ << " (" << data_.rows() << 'x' << data_.cols() << ") to current data set (" << a.rows() << 'x' << a.cols() <<")\n";
    error_abort();
  }
  if (scale_==1.0)
    a+=data_;
  else
    mla(a,scale_,data_);

  checkdomainsw(pflags.back(),pars_.sw,pars_.domain,"direct");
  if (a.rows()>1)
    checkdomainsw(pflags.front(),pars_.sw1,pars_.domain1,"indirect");
}

ProcessCommand* ProcessAdd::create()
{
  const char* fname=parse_string();
  const double scale=are_left() ? parse_double() : 1.0;
  return new ProcessAdd(fname,scale);
}

void ProcessFTIndirect::print(std::ostream& ostr) const
{
  ostr << "ftindirect ";
  print_ftflags(ostr,flags_);
}

ProcessCommand* ProcessFTIndirect::create()
{
  return new ProcessFTIndirect(parse_flags(get_ftflags()));
}

ProcessCommand* ProcessBaselineCorrect::create()
{
  const LIST<size_t> sel(parse_unsignedintarray(1));
  if (sel.empty())
    error_abort("baselinecorrect: column selection cannot be empty");
  return new ProcessBaselineCorrect(sel);
}

void ProcessBaselineCorrect::exec(BaseList<complex> data, processing_state&) const
{
  LIST<complex> sel;
  get_selection(sel,data,inds_);  
  const complex offset=sum(sel)/sel.size();
  data-=offset;
}

Warning<> ProcessSplitSinCos::fd_warning("splitsincos applied to frequency domain data",&NMRsim_once_warning);

void ProcessSplitSinCos::exec(cmatrix& a, BaseList<processing_state> pflags) const
{
  const size_t nrows=a.rows();
  const size_t ncols=a.cols();
  if (ncols & 1)
    error_abort("splitsincos applied to data with odd number of columns");

  if (!(pflags.back().istimedomain))
    fd_warning.raise();

  const size_t newcols=ncols/2;
  cmatrix d(nrows*2,newcols);
  const range realrange(0,newcols-1);
  const range imagrange(newcols,ncols-1);
  for (size_t r=nrows;r--;) {
    BaseList<complex> destrowr(d.row(2*r));
    const BaseList<complex> sourcerow(a.row(r));
    destrowr=sourcerow(realrange);
    BaseList<complex> destrowi(d.row(2*r+1));
    destrowi=sourcerow(imagrange);
  }
  a.swap(d);
  skips.front()=2;
}

void ProcessFTIndirect::exec(cmatrix& a, BaseList<processing_state> pflags) const
{
  const size_t nrows=a.rows();
  const size_t nrows2=nrows/2;
  LIST<complex> tmpr;
  LIST<complex> tmpi(nrows2);
  FTObject obj(nrows,flags_);

  processing_state& curflags=pflags.front();

  if (verbose & VER_GEN)
    std::cout << "Performing ftindirect over " << nrows << " data rows assuming " << ((skips.front()==2) ? "amplitude (States)" : "phase") << " modulated data.\n";

  for (size_t c=a.cols();c--;) {
    switch (skips.front()) {
    case 1: 
      tmpr.create(nrows);
      for (size_t r=nrows;r--;)
	tmpr(r)=a(r,c);
      obj.exec(tmpr,curflags);
      for (size_t r=nrows;r--;)
	a(r,c)=tmpr(r);
      break;

    case 2: {
      tmpr.create(nrows2);
      size_t r,hr;
      for (r=hr=0;hr<nrows;r++,hr+=2) {
	tmpr(r)=complex(a(hr,c).real(),a(hr+1,c).real());
	tmpi(r)=complex(a(hr,c).imag(),a(hr+1,c).imag());
      }
      obj.exec(tmpr,curflags);
      obj.exec(tmpi,curflags);
      for (r=hr=0;hr<nrows;r++,hr+=2) {
	a(hr,c)=complex(tmpr(r).real(),tmpi(r).real());
	a(hr+1,c)=complex(tmpr(r).imag(),tmpi(r).imag());
      }
    }
      break;

    default:
      error_abort("ftindirect - skip can only be 1 or 2");
    }
  }
  curflags.istimedomain=!curflags.istimedomain;
}
	
