#include "Parser.h"
#include "parser_common.hpp"
#include "libcmatrix/ScratchList.h"
#include "libcmatrix/ttyio.h"
#include "expression_definition.hpp"
#ifdef ENABLE_DL
#include <dlfcn.h>
//#include <ltdl.h>
#endif
#include "libcmatrix/cmatrix_hash_set.h"
#include <errno.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <list>
#include <stack>
#include <sstream>
#include <set>

typedef std::set<std::string> envstore_t;
envstore_t envstore; //!< store unique environment variables used - we assume that pointers are unique

namespace {
  const int NMRSIM_DEFAULT_PRECISION=5;
  const int NMRSIM_DEFAULT_MATRIXPRECISION=5;
  const int NMRSIM_DEFAULT_TIMEPRECISION=5;

  const char* getstoredenv(const char* name)
  {
    const char* env=getenv(name);
    if (env)
      envstore.insert(name);
    return env;
  }

}

void dumpenvironment(std::ostream& ostr)
{
  const envstore_t::const_iterator end(envstore.end());
  envstore_t::const_iterator start(envstore.begin());
  while (start!=end) {
    const char* name=(*start++).c_str();
    ostr << name << '=' << getenv(name) << ' ';
  }
}
        
//! line spacer in output .in files
#define OUTPUT_SPACER '\t'

bool isinteractive=false;

BaseWarning NMRsim_repeat_warning(BaseWarning::Always); //!< base for warnings that normally always repeat
BaseWarning NMRsim_once_warning(BaseWarning::FirstOnly); //!< base for warnings that normally are shown just once

ContextWarning<> postcontinuation_whitespace_warning("whitespace after \\ continuation",&NMRsim_repeat_warning);
ContextWarning<> extradollar_warning("variable names should supplied without leading $ (ignored)",&NMRsim_once_warning);
ContextWarning<> namequote_warning("variable names and other string literals ought to be quoted using \'",&NMRsim_once_warning);
ContextWarning<> filetail_ignored_warning("rest of file is being ignored",&NMRsim_repeat_warning);
ContextWarning<> function_redefinition_warning("redefining function: ",&NMRsim_repeat_warning);
//ContextWarning<> misplaced_sum_warning("sum || parameter not meaningful in this context (e.g. in proc rather than initialproc?)",&NMRsim_repeat_warning);
ContextWarning<> misplaced_sum_warning("sum || parameter not meaningful in this context",&NMRsim_repeat_warning);
ContextWarning<> includeonce_withlist_warning("includeonce combined with list arguments.  Doesn't make much sense!",&NMRsim_repeat_warning);
//ContextWarning<> nofunctionarguments_warning("function with no arguments (#<n>): ",&NMRsim_once_warning);
//ContextWarning<> constfunction_warning("function evaluates to a constant! Function: ",&NMRsim_once_warning);
ContextWarning<> xy_variable_warning("using variable names $x or $y may cause confusion with x or y used as phases",&NMRsim_once_warning);
ContextWarning<> trailingcharacters_warning("trailing characters after } are being ignored",&NMRsim_once_warning);
ContextWarning<> ignoreddirective_warning("instruction ignored",&NMRsim_repeat_warning);

dimension_set::incommensurate_warning_t dimension_set::incommensurate_warning("data set size incommensurate with number of previously set rows");
dimension_set::incommensurate_warning_t dimension_set::incommensurate_warning2("list length is incommensurate with previous established length");

#define NMRSIM_OLD_FUNCTION '_'
static char linebuf[MAXLINE];
typedef LIST<std::string> pathlist_t;
pathlist_t pathstore;
varmap_t varmap;
size_t substitute_maxchars=MAXLINE/3;

bool nochecks=false;
bool abortonwarning=false;
bool noexecute=false;
int verbose_level=1;
bool silent=false;

static const double pi=M_PI;
SystemVariable<double*> v_pi("pi",const_cast<double*>(&pi),1.0,V_ISFIXED | V_ISCONST); //const_cast is harmless

static LIST<Variable*> dirty_stack; //!< stack of ::Variable with unset quantity pointers

ContextWarning<> sharedtarget_warning("dirty links have same apparent target",&NMRsim_repeat_warning);

void dirty_stack_push(Variable* varp)
{
  assert(varp!=NULL);
  if ((verbose & VER_GEN) && (verbose_level>1))
    std::cout << "Pushing variable link onto dirty links stack (type=" << (varp->subsid) << ")\n";
  if ((verbose_level>1) && sharedtarget_warning.enabled()) { //!< only check if debugging since arrayed variables may have multiple links to the same target
    subsid_t subsid=varp->subsid;
    for (size_t i=dirty_stack.size();i--;) {
      if (subsid==dirty_stack(i)->subsid) {
	sharedtarget_warning.raise();
	break;
      }
    }
  }
  dirty_stack.push_back(varp);
}

hashed_set<> onceonlyset;
uservarmap_type vardefmap;

optional_map_t& get_optional_map()
{
  static optional_map_t optional_map;
  return optional_map;
}

command_Factory_t& get_Global_Factory() {
  static command_Factory_t Global_Factory;
  return Global_Factory; 
}

void parse_module();
void parse_include(int);
void parse_setenv();

void initialise_precision()
{
  ostream_controller& ctrl(cmatrix_ostream_controller(std::cout));
  std::cout.precision(NMRSIM_DEFAULT_PRECISION);
  std::cerr.precision(NMRSIM_DEFAULT_PRECISION);
  ctrl.matrixprecision=NMRSIM_DEFAULT_MATRIXPRECISION;
  ctrl.timeprecision=NMRSIM_DEFAULT_TIMEPRECISION;
  ctrl.indexbase=1;
}

EXPR_DECLARE_VARIABLE_FUNCTION(FunctionValueof,"Valueof");
EXPR_DECLARE_VARIABLE_FUNCTION(FunctionValuesof,"Valuesof");
EXPR_DECLARE_VARIABLE_FUNCTION(FunctionErrorof,"Errorof");
EXPR_DECLARE_VARIABLE_FUNCTION(FunctionErrorsof,"Errorsof");

namespace {

  struct Parser_Proxy_ {
    Parser_Proxy_() {
      command_Factory_t& Global_Factory(get_Global_Factory());     
      Global_Factory["module"]=par_t(&parse_module,true);
      Global_Factory["includeonce"]=par_t(&parse_include,1,true);
      Global_Factory["include"]=par_t(&parse_include,0,true);
      Global_Factory["variable"]=par_t(&parse_variable,true);
      Global_Factory["function"]=par_t(&parse_function,true);
      Global_Factory["setenv"]=par_t(&parse_setenv,true);
      Global_Factory["verbose"]=par_t(&parse_verbose,true);
      const static FunctionValueof Valueofobj;
      declare_function("Valueof",1U,Valueofobj);
      const static FunctionValuesof Valuesofobj;
      declare_function("Valuesof",1U,Valuesofobj);
      const static FunctionErrorof Errorofobj;
      declare_function("Errorof",1U,Errorofobj);
      const static FunctionErrorsof Errorsofobj;
      declare_function("Errorsof",1U,Errorsofobj);
    }
  };
    
  Parser_Proxy_ parser_proxy_; //!< declare globals

  //! return \c true if \a p is a (readable) directory
  bool isdirectory(const char* p)
  {
    struct stat statbuf;
    
    return stat(p,&statbuf) ? false : S_ISDIR(statbuf.st_mode);
  }

  //! return pointer to continuation point or NULL if none
  char* continuation(char* buf)
  {
    char* p=buf+strlen(buf)-1;
    bool doneinc=false;
    for (;;) {
      if (p<buf)
	return NULL;
      if (!isspace(*p)) {
	if (*p=='\\') {
	  if (doneinc)
	    postcontinuation_whitespace_warning.raise();
	  return p;
	}
	return NULL;
      }
      if ((*p!='\n') && (*p!='\r'))
	doneinc=true; //!< spot any non-CR/LF whitespace
      p--;
    }
  }

  //! return \c true if \a p is an absolute pathname (won't catch DOS style X:\)
  inline bool isabsolute(const char* p)
    {
      return (*p==DIR_SEPARATOR);
    }
}

ContextWarning<> path_unreadable_warning("not a readable directory (skipped): ",&NMRsim_repeat_warning,false);
ContextWarning<> path_emptyelement_warning("empty path element (skipped) in ",&NMRsim_repeat_warning,false);

//! parse path specification
void checkpathname(pathlist_t& pstore, const char* path)
{
  if (!path)
    path=".";
  std::string tmp(path);
  char* p=const_cast<char*>(tmp.c_str()); // a bit naughty - but we are not changing the string size
  char* token;
  while ((token=strtok(p,":"))) {
    p=NULL;
    // don't strip white space?
    if (*token) {
      if (isdirectory(token)) {
	pstore.push_back(token);
	if ((verbose & VER_PARSE) && (verbose_level>1))
	  std::cout << "Adding to path: " << token << '\n';
      }
      else
	path_unreadable_warning.raise(token);      
    }
    else
      path_emptyelement_warning.raise(path);
  }
}

FILE* pathopen(const char* fname, const char* mode)
{
  if (isabsolute(fname))
    return fopen(fname,mode);

  char scr[MAXPATH];
  
  const pathlist_t::const_iterator end(pathstore.end());
  pathlist_t::const_iterator start(pathstore.begin());
  FILE* fp;
  while (start!=end) {
    if (snprintf(scr,MAXPATH,"%s%c%s",start->c_str(),DIR_SEPARATOR,fname)>=MAXPATH)
      throw Failed("pathopen: filename too long");
    fp=fopen(scr,mode);
    if (fp)
      return fp;
    ++start;
  }
  return NULL;
}

void overflowerror()
{
  error_abort("overflow when constructing input line");
}

void pushback(char*& dest, const char* end, char s)
{
  if (dest+1>=end)
    overflowerror();
  *dest++=s;
  *dest='\0';
}

//! return \c true if \a ptr corresponds to a flag argument
/** Flag consists of - followed by only (but at least one) alphanumeric characters 
 */
bool parser_isflag(const char* ptr)
{
  if (!ptr || (*ptr!='-') || isspace(ptr[1]))
    return false;
  while (*(++ptr)) {
    if (isspace(*ptr))
      return true;
    if (!isalpha(*ptr) && (*ptr!='_'))
      return false;
  }
  return true;
}  

//Note "-" doesn't count as flag
bool parser_isflag()
{
  return parser_isflag(get_curline());
}

void pushback(char*& dest, const char* end, const char* source)
{
  const size_t n=strlen(source);
  if (dest+n>=end)
    overflowerror();
  strcpy(dest,source);
  dest+=n;
}

void pushback(char*& dest, const char* end, const char* source, const char* send)
{
  const size_t n=send-source;
  if (dest+n>=end)
    overflowerror();
  memcpy(dest,source,n);
  dest+=n;
}

Mark::Mark()
  : stackpos(dirty_stack.size()) {}

bool Mark::needsflush() const
{
  return (dirty_stack.size()>stackpos);
}

ContextWarning<> Mark::unflushed_warning("mark destroyed before flush (ignore if triggered by exception)",&NMRsim_repeat_warning);

Mark::~Mark()
{
  if (needsflush())
    unflushed_warning.raise();
}

void Mark::flush(Setable* ptr)
{
  if ((verbose & VER_PARSE) && (verbose_level>1))
    std::cout << "Flushing dirty links stack\n";
  assert(dirty_stack.size()>=stackpos);
  while (needsflush()) {
    if (ptr)
      dirty_stack.back()->setptr(ptr);
    dirty_stack.pop_back();
  }
  //  doneflush=true;
}

void checkdirty()
{
  if (!dirty_stack.empty()) {
    dirty_stack.back()->print(std::cerr,false);
    std::cerr << std::endl;
    error_abort("Line finished with dirty objects uncleared");
  }
}

template<> int parse_raw(char*, Variable*, int);
template<> size_t parse_raw(char*, Variable*, int);
void parse_array(LIST<double>&, char*, int flags =F_DENYEXPR);

void Setable::set(const BaseList<double>&, subsid_t subsid)
{
  std::cerr << "Can't set this setable quantity (";
  printvariablename(std::cerr,subsid);
  std::cerr << ") to a list\n";
  error_abort();
}

inline std::ostream& operator<< (std::ostream& ostr, const parse_state& a)
{
  if (a.curline) {
    if (a.lines>1)
      ostr << "lines " << (a.curline-a.lines+1) << '-';
    else
      ostr << "line ";
    ostr << a.curline;
  }
  else
    ostr << "start"; //!< line no. is zero before first line is read
  return ostr << " of " << a.fname;
}

dimension_set::incommensurate_warning_t::incommensurate_warning_t(const char* namev)
  : ContextWarning<>(namev,&NMRsim_repeat_warning) {}
void dimension_set::incommensurate_warning_t::raise(const parse_state& state)
{
  std::ostringstream str(std::ostringstream::out);
  str << " at " << state;
  Warning<>::raise(str.str().c_str());
}

dimension_set array_dims,sum_dims;

struct funcdef
{  
  parse_state state_;
  FILE* fp;
  long fstart;
  size_t startline;
  size_t endline;
  bool active;

  funcdef() : fp(NULL) {}
  funcdef(const char* fname_)
    : state_(fname_), startline(0) { open(fname_); }
  
  funcdef(const char* fname_, long fstart_, size_t startline_, size_t endline_) 
    : state_(fname_), fstart(fstart_), startline(startline_), endline(endline_)
  { open(fname_); }

  ~funcdef();
  
  funcdef* define(const char*);
  void open(const char*);
  bool getline(char* linebuf, size_t max);
  const char* getfname() const { return state_.fname.c_str(); }
  void restart();

  const parse_state& state() const { return state_; }
};

std::ostream& operator<< (std::ostream& ostr, const funcdef& a)
{
  if (a.startline)
    return ostr << a.state_.fname << ':' << a.startline << '-' << a.endline;
  else
    return ostr << a.state_.fname;
}

class frame {
public:
  smartptr<funcdef,false> sourcep;
  LIST<std::string> args;
  std::list<std::string> coms;

  void open(funcdef*, const BaseList<char*>&,bool =false);
  bool skipcomments();
  bool isactive() { return !!sourcep; }
  const char* arg(size_t) const;
  const size_t nargs() const { return getargs().size(); }
  bool isargdefined(size_t n) const { return (n<=nargs()); }
  const parse_state& state() const { return sourcep->state(); }
  void insertline(const char* str) { coms.push_back(str); }
  const char* getfname() const { return sourcep->getfname(); }
  const BaseList<std::string> getargs() { return args; }

private:
  const LIST<std::string>& getargs() const { return args; }
  frame& operator= (const frame&); //disable
};
  

const char* frame::arg(size_t n) const
{
  if (!isargdefined(n)) {
    std::cerr << state() << ": invalid argument number: " << n << '\n';
    error_abort();
  }
  return getargs()(n-1).c_str();
}

frame* curframep(NULL);
frame* topframep(NULL);
std::string inname;

const char* parser_getfname()
{
  return inname.c_str();
}

double curgrat=0.0;
double defgrat=0.0; //!< set to unique nucleus grat for default

static const double rad_to_deg=180.0/M_PI;

template<> double parse_raw(char*, Variable*, int);

class Strpbrk {
public:
  Strpbrk(const char*);
  char* operator()(char*) const;
private:
  ScratchList<bool,256> tableok_;
};

Strpbrk::Strpbrk(const char* matchchars)
  : tableok_( sizeof(char) << 8, true)
{
  while (*matchchars) {
    tableok_(size_t(*matchchars))=false;
    matchchars++;
  }
  tableok_(size_t(0))=false; //!< convenient if also stop on '\0';
}
 
char* Strpbrk::operator()(char* p) const
{
  for (;tableok_(*p);p++);
  return *p ? p : NULL;
}

char matchingbracket(char start)
{
  static char matchtable[sizeof(char)<<8]={'\0'};
  static bool havetable=false;
  if (!havetable) {
    matchtable[size_t('{')]='}';
    matchtable[size_t('}')]='X';
    matchtable[size_t('(')]=')';
    matchtable[size_t(')')]='X';
    matchtable[size_t('[')]=']';
    matchtable[size_t(']')]='X';
    matchtable[size_t('\'')]='\'';
    matchtable[size_t('"')]='"';
    matchtable[size_t('`')]='`';
    matchtable[size_t('|')]='|';
  }
//   switch (start) {
//   case '{':
//     return '}';
//   case '(':
//     return ')';
//     break;
//   case '[':
//     return ']';
//   case '"':
//     return '"';
//   case '\'':
//     return '\'';
//   }
  const char res(matchtable[size_t(start)]);
  switch (res) {
  case 'X':
    return '\0';
  case '\0':
    throw InternalError("matchingbracket");
  }
  return res;
}

bool checkbracket(char*& cptr, char start)
{
  if (*cptr!=start)
    return false;
  const char end=matchingbracket(start);
  char* last=cptr+strlen(cptr+1);
  if (*last!=end) {
    parser_printcontext() << "Failed to find matching " << end << '\n';
    error_abort();
  }
  cptr++;
  *last='\0';
  return true;
}

VariableBase* parse_double_variable(Variable& cvar, int flags)
{
  (void)parse_double_raw(&cvar,get_token(),flags | F_ALWAYSNEW);
  return cvar.valuep;
}

const char* clean_variable_name(char* name)
{
  if (*name=='$') { 
    extradollar_warning.raise();
    name++;
  }
  else {
    if (!checkbracket(name,'\''))
      namequote_warning.raise();
  }
  return name;
}

RealVariable::RealVariable(const char* namev, VariableBase& varv)
  : name_(namev), varp_(&varv), issetable_(true), isconst_(varv.isconst()), isused_(false)
{
  if (!namev)
    throw InvalidParameter("RealVariable");
}  

RealVariable& parse_variable_name(const char* cptr, bool flagused)
{
  const systemvarmap_type::const_iterator curp(systemvarmap.find(cptr));
  if (curp!=systemvarmap.end()) {
    RealVariable* gptr=dynamic_cast<RealVariable*>(curp->second);
    if (!gptr) {
      parser_printcontext() << '$' << cptr << " does not resolve to a numeric variable" << std::endl;
      throw Failed("parse_variable_name");
    }
    if (flagused)
      gptr->flagused();
    return *gptr;
  }
  return *findvariable(cptr,flagused);
}

void parse_array(LIST<double>& tmparray, char* cptr, int flags)
{
  if (!checkbracket(cptr,'{') && !checkbracket(cptr,'[')) {
    parser_printcontext() << "Expected {} or [] delimited list while parsing: " << cptr << '\n';
    error_abort();
  }
  chewwhite(cptr);
  if (*cptr=='\0') {
    tmparray.clear();
    return;
  }
    
  expr_grammar& expr_parser(get_expr_parser());
  expr_parser.reset(flags | F_ALLOWLIST);
  const parse_info<> info(parse(cptr,
			  expr_parser.use_parser<expr_grammar::rawlist_def>(),
			  space_p));
  if (info.full) {
    VariableBase& var(expr_parser.variable());
    if (!var.isconst())
      error_abort("expression does not evaluate to constant list\n");
    tmparray.swap(var.value());
    return;
  }
  parser_printcontext() << "list parsing failed at: \"" << info.stop << "\"\n";
  error_abort();
}

void fudge_shift(double& val, size_t qualifier)
{
  val*=get_nmrfreq(qualifier)*1e-6;
}

void register_expression(const Variable& var)
{
  Variable& link=push_and_get(expressions,var);
  if (verbose & VER_PARSE)
    std::cout << "Created expression: " << *(var.valuep) << '\n';
  if (!link.ptr)
    dirty_stack_push(&link);
}

VarVariable::VarVariable(const Variable& varv, Parameter& parv, int arraywhichv)
  :
  Variable(varv), 
  arraywhich((arraywhichv==-1) ? 0U : arraywhichv),
  parameter(parv)
{
  const double step=variable().get_error(arraywhich);
  if (step==0.0)
    throw InternalError("VarVariable created from fixed parameter");
  parameter.error(step);
  //  updatebound();
}

// void VarVariable::updatebound()
// {  
//   const BaseBoundFunction* funcp=boundstate.function();
//   if (funcp==NULL) {
//     parameter.unconstrain();
//     return;
//   }
//   const double val=get();
//   if (!(funcp->isvalid(val))) {
//     parser_printcontext() << "Current value (" << val << ") of parameter " << parameter.name() << " is outside bound range: " << boundstate << '\n';
//     error_abort();
//   }
//   parameter.constrain(*funcp);
// }

double handle_variable(Variable* vardefp, int flags)
{
  VariableBase* newvarp=NULL;
  bool found=false;
  if (vardefp && vardefp->ptr) {
    varmap_t::iterator iter=varmap.find(vardefp->key());
    if (iter!=varmap.end()) {
      found=true;
      if (!(flags & F_ALWAYSNEW))
	throw InternalError("Redefine demands ALWAYSNEW");
      newvarp=iter->second;
      if (!(newvarp->isconst()))
	error_abort("Can only redefine const variable");
    }
  }
  expr_grammar& expr_parser(get_expr_parser());
  VariableBase& cvar(expr_parser.variable());
  const BaseList<double> steps(cvar.get_errors());
  const double value=cvar.value().size() ? cvar.value().front() : 0.0; //0 if empty array

  if (cvar.issimple() && (cvar.value().size()<2) && steps.empty() && !(flags & F_ALWAYSNEW)) {
    //    if (cvar.value().size()>2) //!< allow 0 or 1 element
    //  throw InternalError("handle_variable:2");
    if (vardefp) {
      vardefp->subsid=S_NONE;
      vardefp->valuep=NULL;
    }
    expr_parser.builder_.reset(); //reset parser so another variable can be parsed
    return value;
  }
  if (!vardefp || (vardefp->subsid==S_NONE))
    error_abort("Can't use non-constant value here");

  if (newvarp) {
    if (!(cvar.isconst()))
      error_abort("Can only overwrite existing variable with constant expression");
  }
  else
    newvarp=new VariableBase();

  newvarp->swapin(cvar); //swap in contents
  vardefp->valuep=newvarp;

  if (newvarp->isexpr()) {
    register_expression(*vardefp);
    if (!steps.empty())
      throw InternalError("handle_variable");
  }
  else {
    if (newvarp->isarray()) { //'register' array sizes      
      const ListList<double> arrayvals(newvarp->get_array());
      if ((arrayvals.size()>1) && (flags & F_DENYSUM))
	misplaced_sum_warning.raise();
      verify_arraysizes(arrayvals,newvarp->sumtag(),newvarp->arraytags());
      Variable& link=push_and_get(varpars,*vardefp);
      if (!(link.ptr))
	dirty_stack_push(&link);
      //check for fitted elements
      for (size_t i=0;i<steps.size();i++) {
	if (steps(i)) {
	  //	  dirty_stack_push(create_fitting_variable(VarVariable(link,steps(i),i)));
	  parameter_list.push_back(Parameter(arrayvals.row()(i),steps(i)));
	  register_fitting_variable(new VarVariable(link,parameter_list.back(),i));
	}
      }
    }
    else {
      double err=0.0;
      switch (steps.size()) {
      case 0:
	break;
      case 1: 
	err=steps.front();
	break;
      default:
	throw InternalError("handle_variable");
      }
      if (err) {
	parameter_list.push_back(Parameter(newvarp->value().front(),steps.front()));
	register_fitting_variable(new VarVariable(*vardefp,parameter_list.back()));
      }
      else {
	if (newvarp->isconst())
	  vardefp->subsid=S_NONE; // flag not variable      
	//! NB does not remove any previous Variable from map
      }
    }
  }
  expr_parser.builder_.reset(); //reset parser so another variable can be parsed

  return value;
}

double handle_variable(int flags, size_t subsid)
{
  Variable cvar(subsid_t(subsid),NULL);
  return handle_variable(&cvar,flags);
}

double parse_double_raw(Variable* vardefp, char* cptr, int flags)
{
  if (!vardefp && (flags & F_ALWAYSNEW))
    throw InternalError("parse_double_raw");

  expr_grammar& expr_parser(get_expr_parser());
  expr_parser.reset(flags);
  const parse_info<> info(parse(cptr,
			  expr_parser.use_parser<expr_grammar::variable_def>(),
			  space_p));
  if (!info.full) {
    parser_printcontext() << "Parsing failed at: \"" << info.stop << "\"\n";
    error_abort();
  }  
  return handle_variable(vardefp,flags);
}
  
double parse_shift(Variable* vardef, double grat, int flags)
{
  curgrat=grat;
  //  double retval=parse_double(vardef,flags | F_ISSHIFT);
  double retval=parse_double(vardef,flags);
  curgrat=defgrat; //unset to trap attempts to parse 'p' without defining nucleus
  return retval;
}

template<> double parse_raw(char* cptr, Variable* cvarp, int flags)
{
  return parse_double_raw(cvarp,cptr,flags);
}

double parse_double(Variable* vardef, int flags)
{
  return parse_raw<double>(get_token(),vardef,flags);
}

size_t parser_verbose_level()
{
  return (verbose & VER_PARSE) ? verbose_level : 0;
}

funcdef::~funcdef()
{
  if (parser_verbose_level()>1)
    std::cout << "Closing frame: " << state_.fname << '\n';
  if ((fp!=stdin) && fp)
    fclose(fp);
}

void funcdef::open(const char* fname_)
{
  static const char stdinname[]="<stdin>";
  if (strcmp(fname_,"-")==0) {
    state_.fname=stdinname;
    fp=stdin;
  }
  else {//postpone opening file until required
    state_.fname=fname_;
    fp=NULL;
  }
  active=false;
}

void parser_insertline(const char* buf)
{
  if (verbose & VER_PARSE)
    std::cout << "Pushing line onto input stack: " << buf << '\n';
  curframep->insertline(buf);
}

bool funcdef::getline(char* linebuf, size_t max)
{
  if (!active)
    return false;

  const char* cptr=NULL;
  const bool usestack=!(curframep->coms.empty());
  size_t lines=0;
  if (usestack) {
    cptr=curframep->coms.front().c_str();
    lines=1;
  }
  else {
    cptr=linebuf;
    char* dest=linebuf;
    int left=max;   
    while (active) {
      if (fgets(dest,left,fp)) {
	lines++;
	state_.curline++;
	char* contp=continuation(dest);
	if (state_.curline+1==endline) {//reached end of parse lines?
	  if (contp)
	    error_abort("continuation \\ at end of include block");
	  active=false;
	}
	else {
	  if (contp) {
	    left-=contp-dest;
	    assert(left>=0);
	    dest=contp;
	  }
	  else
	    break;
	} 
      }
      else
	active=false;
    }
    if (!active && (verbose & VER_PARSE))
      std::cout << "End of frame: " << (*this) << '\n';
  }

  if (lines==0) { // Failed to find input line
    active=false;
    return false;
  }
  state_.lines=lines;

  const size_t n=strlen(cptr);
  if (n>=max-1) {
    parser_printcontext() << "input line too long (max of " << max << " chars)\n";
    error_abort();
  }
  if (usestack) {
    strcpy(linebuf,cptr);
    curframep->coms.pop_front();
  }

  if (n) { // if non-zero string, strip terminating white space
    char* endchar=linebuf+n-1;
    while (isspace(*endchar))
      endchar--;
    endchar[1]='\0';
  }
  return true;
}

funcdef* funcdef::define(const char* blockname)
{
  char linebuf[MAXLINE];
  const long newfstart=ftell(fp);
  const size_t newstartline=state_.curline+1;
  size_t newendline;
  for (;;) {
    if (!fgets(linebuf,sizeof(linebuf),fp)) {
      parser_printcontext() << "Exhausted file when looking for end of " << blockname << '\n';
      error_abort();
    }
    if (continuation(linebuf))
      error_abort("\\ continuation not allowed in block definition");
    state_.curline++;
    char* cptr=linebuf;
    chewwhite(cptr);
    if (*cptr=='}') {
      newendline=state_.curline;
      break;
    }
  }
  funcdef* newfuncp= new funcdef(state_.fname.c_str(),newfstart,newstartline,newendline);
  if (verbose & VER_PARSE)
    std::cout << blockname << " defined: " << (*newfuncp) << " at file pointer: " << newfstart << '\n';
  return newfuncp;
}

void funcdef::restart()
{
  if (active)
    throw Failed("Recursive file open");
  if (!fp) {
    fp=pathopen(state_.fname.c_str(),"r");
    if (!fp) {
      parser_printcontext() << "failed to open " << state_.fname << std::endl;
      error_abort(ERR_FAILED);
    }
  }
  if (startline)
    fseek(fp,fstart,SEEK_SET);
  active=true;
  state_.curline=startline ? startline-1 : 0;
}

typedef MAPTYPE(funcdef*) funcdefmap_type;
funcdefmap_type funcdefmap;

void declare_builtin_block(const char* name)
{
  funcdefmap[name]=NULL;
}

void par_t::operator()()
{
  if (!allowmultiple) {
    if (done)
      error_abort("Can't reuse this directive");

    done=true;
  }
  if (parsefunc)
    (*parsefunc)();
  else {
    if (parsefunc_qual)
      (*parsefunc_qual)(ident);
    else
      throw InternalError("par_t");
  }
}

Splitter::Splitter(char* source, const char* charsv, bool uniquev)
  : curptr(source), chars(charsv), unique(uniquev)
{ 
  sepchar='\0';
  chewwhite(curptr);
}
  
char* Splitter::next(char& retsepchar)
{
  if (atend())
    return NULL;

  retsepchar=sepchar;

  s=NULL;
  for (const char* trychar=chars;*trychar;trychar++) {
    char *cur(strchr((*trychar=='-') ? curptr+1 : curptr,*trychar));
    if ( cur && ((cur<s) || !s))
      s=cur;
  }
  char* end;
  if (s) {
    sepchar=*s;
    end=s-1;
    if (unique && retsepchar && (sepchar!=retsepchar)) {
      parser_printcontext() << "cannot mix item separators: changed from " << retsepchar << " to " << sepchar  << '\n';
      error_abort();
    }
  }
  else {
    end=curptr+strlen(curptr)-1;
    sepchar='\0';
  }

  while (isspace(*end))
    end--;
  end[1]='\0'; //remove trailing spaces

  char* oldptr=curptr;
  if (!*(oldptr)) //separator at start of string?
    throw Failed("Couldn't parse expression");

  if (sepchar) {
    curptr=s+1;
    chewwhite(curptr);
  }
  else
    curptr=NULL;
  return oldptr;
}

//Parser state
const char* blockname=NULL;
bool insection=false;
bool needfresh=false;
bool lastblockwasdefinition=false;
bool found_block=false;
bool have_block=false;
const char** keynames;
size_t nkeys;
bool isopt;
std::stack< frame, LIST<frame> > fps;

static bool popinclude();

bool parser_foundlast() { return found_block; }
bool parser_inblock() { return insection; }

int global_flags=0;

void parser_init(const char* fname, int argc, char* argv[],int global_flagsv)
{
  BaseList<char*> args(argc,argv);
  checkpathname(pathstore,getenv("NMRSIM_PATH"));
  fps.push(frame());
  inname=fname;
  topframep=curframep=&(fps.top());  
  curframep->open(new funcdef(fname),args);
  global_flags=global_flagsv;

  add_systemvarmap(v_pi);
  initialise_precision();
}

bool parser_isopen()
{
  return curframep ? curframep->isactive() : false;
}

const char delims[3]={ ' ','\t','\0' };
static const double deg_to_rad=M_PI/180.0;

static char* placeholder=NULL;
static char* lasttoken=NULL;

#ifndef HAVE_STRTOK_R
// replacement for strtok_r if missing
char* strtok_r( char *str, const char *delim, char **save_ptr ) 
{
  if (!str)
    str=*save_ptr;

  str+=strspn(str,delim); //! skip leading delimiters

  if (!*str) {
    *save_ptr=str;
    return NULL;
  }

  char* ptr = str;

  while (*ptr && !strchr(delim,*ptr))
    ++ptr;

  *save_ptr = *ptr ? ptr + 1 : ptr;
  *ptr = '\0';

  return str;
}
#endif

ContextWarning<> redundantquote_warning("Quoting arguments with \" is no longer necessary",&NMRsim_once_warning);

char* token_search(char*& cptr, char endpar ='\0', bool optional =false)
{
  static const char paras[]="\"\'`(){}[]|";
  static const Strpbrk strpbrk_paraset(paras); //!< use cached strpbrk for larger number of characters
  static LIST<char> parstack(size_t(10));  
  static char pdelims[2]={'\0'};

  const char* ldelims=delims;
  bool ispar=false;
  if (endpar) {
    if (strchr(paras,endpar)) {
      if (optional)
	throw InternalError("token_search: optional token separator can't be parenthesis/quote");
      ispar=true;
      parstack.create(1,endpar);
    }
    pdelims[0]=endpar;
    ldelims=pdelims;
  }
  else
    parstack.create(size_t(0)); //reset paren stack

  char* lptr=cptr;
  char* nextdelimp=NULL;
  bool stripquote=false;
  char insideliteral='\0';
  for (;;) {
    if (lptr>nextdelimp) //!< need to search for end delimiter?
      nextdelimp=strpbrk(lptr,ldelims); 
    if (!nextdelimp) {
      if (endpar) {
	if (optional)
	  return NULL;
	parser_printcontext() << "Failed to find expected " << endpar << '\n';
	error_abort();
      }
      nextdelimp=lptr+strlen(lptr);
    }
    char* nextpar=NULL;
    if (insideliteral) {
      for (;;) {
	nextpar=strchr(lptr,insideliteral);
	if ((nextpar<=lptr) || (nextpar[-1]!='\\')) //not quoted?
	  break;
	lptr=nextpar+1;
      }
    }
    else {
      const bool shortcircuit=parstack.empty();
      char delimchar='\0';
      if (shortcircuit) {
	delimchar=*nextdelimp;
	*nextdelimp='\0';
      }
      nextpar=strpbrk_paraset(lptr);
      if (shortcircuit)
	*nextdelimp=delimchar; //!< restore
    }
    
    if (nextpar==NULL)
      break;
    if (insideliteral)
      insideliteral='\0';
    else {
      bool islit=false;
      switch (*nextpar) {
      case '\'': case '`':
	islit=true;
	break;
      case '\"':
	if ((!endpar) && (nextpar==cptr)) {
	  stripquote=true;
	  redundantquote_warning.raise();
	}
	else {
	  if (!(stripquote && (nextpar==nextdelimp-1)))
	    islit=true;
	}
	break;
      }
      if (islit)
	insideliteral=*nextpar;
      else {
	if (!(parstack.empty()) && (*nextpar==parstack.back())) {
	  parstack.pop_back();
	  if (*nextpar==endpar && (parstack.empty()))
	    return nextpar;
	}
	else {
	  const char match=matchingbracket(*nextpar);
	  if (!match) {
	    parser_printcontext() << "Unexpected closing parenthesis (" << *nextpar << ')';
	    if (!(parstack.empty()))
	      std::cerr << ". Expecting " << parstack.back();
	    std::cerr << " when parsing " << cptr << '\n';
	    error_abort();
	  }
	  parstack.push_back(match);
	}
      }
    }
    lptr=nextpar+1;
  }
  if (parstack.size()) {
    parser_printcontext() << "unterminated quote: missing trailing " << parstack.back() << '\n';
    error_abort();
  }
  if (stripquote) {
    if (nextdelimp[-1]!='\"')
      error_abort("argument quoted with \" is not isolated phrase");
    nextdelimp[-1]='\0';
    cptr++;
  }
  return nextdelimp;
}

char* get_token_raw(char*& holder, int flags =0)
{
  char* cptr=holder;
  if (cptr) {
    chewwhite(cptr);
    char* nextdelimp=token_search(cptr);
    if (*nextdelimp) {
      *nextdelimp='\0';
      holder=nextdelimp+1;
    }
    else 
      holder=NULL;
    
//     if (*cptr=='\"') {
//       holder=strchr(++cptr,'\"');
//       if (!holder)
// 	error_abort("Unbalanced \"");
//       *holder++='\0';
//     }
//     else
//       cptr=strtok_r(NULL,delims,&holder);
  }
  if ((cptr==NULL) && !(flags & F_ALLOWMISSING))
    error_abort("Missing argument");
  return cptr;
}

char* get_token(int flags)
{
  lasttoken=get_token_raw(placeholder,flags);
  return lasttoken;
}

bool are_left()
{
  if (!placeholder)
    return false;
  chewwhite(placeholder);
  return (*placeholder!='\0');
}

size_t count_left(bool countflags)
{
  char* inpargs=get_curline();
  if (!inpargs)
    return 0;
  ScratchList<char> copyargs( BaseList<char>( 1+strlen(inpargs) ,inpargs) );
  //need to copy arguments as strtok will break up...
  char* localholder=copyargs.vector();
  size_t count=0;
  char* ptr;
  while ((ptr=get_token_raw(localholder,F_ALLOWMISSING))) {
    if (countflags || !parser_isflag(ptr))
      count++;
  }
  return count;
}

void parser_newblock(const char* blockname_, bool isopt_, const char* keynames_[],size_t nkeys_)
{
  if (insection)
    throw Failed("parser_newblock: can't start new block inside active block"); 
  blockname=blockname_;
  keynames=keynames_;
  nkeys=nkeys_;
  isopt=isopt_;
  found_block=false;
}

bool popinclude()
{
  fps.pop();
  if (fps.empty()) {
    curframep=topframep=NULL; //finished all files?
    return false;
  }
  curframep=&(fps.top());
  return true;
}

void frame::open(funcdef* sourcep_, const BaseList<char*>& largs, bool persist)
{
  if (!largs.empty())
    args=largs; //store arguments

  sourcep.reset(sourcep_,persist ? mxflag::nondynamic : mxflag::normal);
  if (verbose & VER_PARSE) {
    std::cout << "Opening " << (*sourcep);
    if (args.size()) {
      std::cout << " with arg list: ";
      for (size_t i=0;i<args.size();i++) {
	const char* arg(args(i).c_str());
	if (*arg)
	  std::cout << arg << ' ';
	else
	  std::cout << "'' ";
      }
    }
    std::cout << '\n';
  }
  if (sourcep->active) {
    parser_printcontext() << "Recursive call to " << (*sourcep) << '\n';
    error_abort();
  }
  sourcep->restart();
}

bool frame::skipcomments()
{
  lasttoken=NULL; //!< clear references to previous lines
  char rawlinebuf[MAXLINE];
  while (sourcep->getline(rawlinebuf,MAXLINE)) {
    substitute_string(linebuf,sizeof(linebuf),rawlinebuf,SUB_NUMERIC | SUB_ABORTHASH);
    if (linebuf[0]!='#')
      return true;
  }
  return false;
}

void parser_getinteractive(const char* intro, std::istream& istr)
{
  char rawlinebuf[MAXLINE];
  for (;;) {
    std::cout << intro;
    getline(rawlinebuf,MAXLINE);
    substitute_string(linebuf,sizeof(linebuf),rawlinebuf,SUB_NUMERIC | SUB_ABORTHASH);
    if (*linebuf) {
      set_curline(linebuf);
      return;
    }
  }
}

ContextWarning<> lastblockdefinition_warning("final block was parsed as an internal include - mislabelled block?  Parser was expecting: ",&NMRsim_repeat_warning); 

void parser_flush()
{
  if (insection || !needfresh)
    error_abort("file incompletely parsed (use -verbose:parse to diagnose)");
  if (lastblockwasdefinition)
    lastblockdefinition_warning.raise(blockname);
  for (;;) {
    if (!curframep)
      return;
    if (curframep->skipcomments())
      break; //!< oh dear, found more text
    if (!popinclude())
      return; //!< OK, nothing left
  }
  filetail_ignored_warning.raise();
}

void parse_module()
{
  if (have_block)
    error_abort("Modules must be loaded at start");
#ifndef ENABLE_DL
  error_abort("Dynamic module load not supported");
#else
  static LIST<std::string> loaded_modules;
  if (!are_left()) {
    std::cout << "Modules loaded:";
    for (size_t i=0;i<loaded_modules.size();i++)
      std::cout << ' ' << loaded_modules(i);
    std::cout << '\n';
    return;
  }
//   if (lt_dlinit()) {
//     std::cerr << "Failed to initialise libltdl\n";
//     exit(1);
//   }
  for (;;) {
    char* fname=parse_string(F_ALLOWMISSING);
    if (!fname)
      return;
    void* handle=dlopen(fname,RTLD_LAZY);
    //lt_dlhandle handle=lt_dlopenext(fname);
    if (!handle) {
      std::cerr << "Failed to load module: " << fname << ": " << dlerror() << '\n';
      error_abort();
    }
    else
      loaded_modules.push_back(getbasename(fname));
  }
#endif
}

namespace {
  struct tmparg_t {
    char str[10];
    char* operator()() { return str; }
    const char* operator()() const { return str; }
    tmparg_t() { *str='\0'; }
  };
}
    
void parse_include(int onceonly)
{
  const char* fname(parse_string());

  LIST<char*> args;
  LIST< BaseList<double> > varargs;
  LIST<tmparg_t> tmpargs;
  size_t maxrep=1;
  while (are_left()) {
    char* tokptr=get_token();
    const bool isquoted=checkbracket(tokptr,'\'');   
    BaseList<double> argptr;
    if (!isquoted) {
      Variable cvar(S_ARG1);
      const double v=parse_double_raw(&cvar,tokptr,F_DENYVAR | F_DENYARRAY | F_ALLOWLIST | F_ALWAYSNEW);
      assert(cvar.subsid==S_NONE);
      tmpargs.push_back(tmparg_t()); //create new buffer space
      tokptr=tmpargs.back()();
      argptr.create(cvar.valuep->value());
      const size_t n=argptr.size();
      if (n==1)
	snprintf_prec(tokptr,sizeof(tmparg_t),v);
      else {
	if ((n!=1) && (n!=maxrep)) {
	  if (maxrep==1)
	    maxrep=n;
	  else
	    throw Failed("include: arguments have incompatible lengths");
	}
      }
    }      
    args.push_back(tokptr);
    varargs.push_back(argptr);
  }

  if (onceonly) {
    if (maxrep!=1)
      includeonce_withlist_warning.raise();

    const bool already=(onceonlyset.count(fname)!=0);
    if (verbose & VER_PARSE)
      std::cout << "File " << fname << " already included: " << (already ? "Yes\n" : "No\n");
    if (already)
      return; //already included
    onceonlyset.insert(fname);
  }
    
  //! empty list?
  if (!maxrep)
    return;

  static char tlinebuf[MAXLINE]="include XX"; // XXX is fence-post check
  if (maxrep>1) {
    char* const endptr=tlinebuf+sizeof(tlinebuf);
    for (size_t i=0;i<maxrep;i++) {
      char* curptr=tlinebuf+sizeof("include ")-1;
      pushback(curptr,endptr,fname);
      for (size_t j=0;j<varargs.size();j++) {
	const BaseList<double>& largs(varargs(j));
	char* tokptr=args(j);
	if (largs.size()>1) //!< need to change argument?
	  snprintf_prec(tokptr,sizeof(tmparg_t),largs(i)); //!< re-format numerical argument
	pushback(curptr,endptr," \'"); //!< inter-argument space
	pushback(curptr,endptr,tokptr); //!< add argument
	pushback(curptr,endptr,'\''); //!< need to quote to prevent further evaluation
      }
      parser_insertline(tlinebuf);
    }
    return;
  }
    
  fps.push(frame());
  frame& newframe=fps.top();
  const funcdefmap_type::const_iterator curp(funcdefmap.find(fname));
  const bool newfile(curp==funcdefmap.end());
  funcdef* sourcep=newfile ? new funcdef(fname) : curp->second;
  if (!sourcep) {
    parser_printcontext() << "Attempt to include built-in block " << fname << '\n';
    error_abort();
  }
  newframe.open(sourcep,args,!newfile);
  curframep=&newframe;
}

#define CHECKWORD(A,B)\
  ((STRNCMP(A,B)==0) && (isspace(cptr[sizeof(B)-1]) || !cptr[sizeof(B)-1]))


bool parser_getline(bool executeglobal)
{
  if (!blockname)
    error_abort("parser_getline: no block started");
  if (!curframep)
    return false;
  command_Factory_t& Global_Factory(get_Global_Factory());
  for (;;) {
    if (needfresh && !(curframep->skipcomments())) {
      if (!popinclude())
	return false;
      continue;
    }
    needfresh=true;
    char* cptr=linebuf;
    chewwhite(cptr);
    if (*cptr=='}') {
      if (!insection)
	error_abort("Unexpected } block end");
      cptr++;
      chewwhite(cptr);
      if (*cptr)
	trailingcharacters_warning.raise();
      insection=false;
      if (verbose & VER_PARSE)
	parser_printcontext() << "finishing parsing block: " << blockname << '\n';
      return false; // end block
    }
    if (!*cptr)
      continue;

    if (verbose & VER_PARSE)
      std::cout << "Parsing " << (curframep->state()) << ": " << cptr << std::endl;

    char* tail=cptr+1;
    while (*tail && !isspace(*tail))
      tail++;
    const char storechar=*tail;
    *tail='\0';

//     if (strcmp(cptr,"include")==0) {
//       *tail=storechar;
//       set_curline(tail);
//       parse_include(false);
//       continue;
//     }
//     if (strcmp(cptr,"includeonce")==0) {
//       *tail=storechar;
//       set_curline(tail);
//       parse_include(true);
//       continue;
//     }
//    const bool isglobal=Global_Factory.count(cptr);
    command_Factory_t::iterator curp=Global_Factory.find(cptr);
    const bool isglobal=(curp!=Global_Factory.end());
    if (isglobal) {
      set_curline(storechar ? tail+1 : tail);
      if ((curp->second).parsefunc==parse_setenv) { //!< always execute setenv
	parse_setenv(); //!< execute setenv *AND* return line if in scan mode
	continue;
      }
      else {
	const bool isinclude=((curp->second).parsefunc_qual==parse_include);
	if (isinclude || executeglobal) {
	  (curp->second)();
	  parser_checkfinished();
	  continue;
	}
      }
    }
    *tail=storechar;
    set_curline(cptr);
//     if (CHECKWORD(cptr,"include")) {
//       set_curline(cptr+sizeof("include"));
//       parse_include(false);
//       continue;
//     }
//     if (CHECKWORD(cptr,"includeonce")) {
//       set_curline(cptr+sizeof("includeonce"));
//       parse_include(true);
//       continue;
//     }
    if (!isglobal && !insection) {
      char tmp[128];
      char tmpc;
      if ((sscanf(cptr,"%s %c ",tmp,&tmpc)!=2) || (tmpc!='{'))
	error_abort("Expecting block start");
      if (strcmp(tmp,blockname)==0) {
	have_block=true;
	found_block=true;
	insection=true;
	lastblockwasdefinition=false;
	if (verbose & VER_PARSE)
	  parser_printcontext() << "found expected parsing block: " << blockname << '\n';
	continue;
      }
      const funcdefmap_type::const_iterator curp(funcdefmap.find(tmp));
      if (curp!=funcdefmap.end()) {
	if (curp->second)
	  function_redefinition_warning.raise(tmp);
	else {
	  if (isopt) {
	    have_block=true;
	    needfresh=false;
	    return false; //found different block
	  }
	  parser_printcontext() << "can't redefine built-in block (out of order?): " << tmp << std::endl;
	  error_abort();
	}
      }
      //define new block
      funcdefmap[tmp]=curframep->sourcep->define(tmp);
      lastblockwasdefinition=true;
      continue;
    }
    return true;
  }
}

char* get_curline() 
{
  char* ptr=placeholder;
  if (ptr)
    chewwhite(ptr);
  return ptr;
}

//bit dodgy
void set_curline(char* ptr) {
  placeholder=ptr;
}

//! return \c true if \a str does not contain lower case characters
bool isenvironment(const char* str)
{
  while (*str) {
    if (islower(*str))
      return false;
    str++;
  }
  return true;
}

inline char* strchrlimit(char* s, char* tail, char c)
{
  char* p=strchr(s,c);
  return (!p || (p>=tail)) ? NULL : p;
}

namespace {

  bool isvarconstant(const RealVariable* vp) { return vp->isconstant(); }

  bool isvarconstant(const SystemVariableBase* vp)
  {
    const RealVariable* vr=dynamic_cast<const RealVariable*>(vp);
    return vr ? isvarconstant(vr) : true;
  }
};

Warning<> nonconstsubs_warning("Formatting non-constant variable outside of main execution loop at $",&NMRsim_once_warning);

char* substitute_string(char* buffer, int nchars, const char* rawcptr, int flags)
{
  char* cptr=const_cast<char*>(rawcptr); //!< a bit of a hack, but we only tweak the input string and don't fundamentally violate its constness
  char* curptr=buffer;
  char* endptr=buffer+nchars;
  char tmp[32];

  *buffer='\0';
  if (!cptr)
    return buffer;

  while (cptr && *cptr) {
    char* dollar=strchr(cptr,'$');
    if (!dollar) {
      pushback(curptr,endptr,cptr);
      return curptr;
    }
    // skip over additional $
    while (dollar[1]=='$')
      dollar++;

    pushback(curptr,endptr,cptr,dollar);

    dollar++; //point to post$

    if ((curptr!=buffer) && (curptr[-1]=='\\')) { //found \$
      if (flags & SUB_ESCAPE)
	curptr--; //move back to write over
      *curptr++='$';
      cptr=dollar;
      continue;
    }
    bool hasparen=false;
    int skip=0;
    char* tail=NULL;
    if (*dollar=='(') {
      hasparen=true;
//       tail=strchr(++dollar,')');
//       if (!tail)
// 	error_abort("Unbalanced () in $(<variable>)");
      dollar++;
      tail=token_search(dollar,')');
      if (!tail)
	throw InternalError("token_search");
      skip=1;
    }

    const bool isstar=(*dollar=='*');
    if (isstar)
      dollar++;
    char* origcptr=dollar;

    long n=-1;
    if (flags & SUB_NUMERIC) {
      char* ltail;
      n=strtoul(dollar,&ltail,10);
      if (ltail==dollar)
	n=isstar ? 1 : -1; // flag invalid unless $* (equivalent to $1)
      if ((n>=0) && !hasparen)
	tail=ltail;
    }
    // establish end of name
    if (n<0) { //didn't find digit
      if (isstar)
	error_abort("Can only use $* with argument variables e.g. $*1");
      if (!tail) {//!< unless end already known, look for white space or (
	char* endptr=dollar;
	while (isalpha(*endptr) || isdigit(*endptr) || (*endptr=='_'))
	  endptr++;
	if (*endptr)
	  tail=endptr;
      }

      if (tail) {
	char* oldptr=curptr;
	pushback(curptr,endptr,dollar,tail); //copy into buffer (scratch copy)
	*curptr='\0'; //null terminate
	dollar=curptr=oldptr; //reset ptr;
	if (!(flags & SUB_NUMERIC))
	  tail+=skip; //!< done later for substitution variables (yuk)
      }      
      // catch bad variable name e.g. $.
      if (*dollar=='\0')
	error_abort("invalid use of $: following variable name can contain only alphanumeric characters (letters and numbers)");
    }
    char* qualifier=NULL;
    char* haspositive=NULL;
    if (hasparen) {            
      qualifier=(n<0) ? strchr(dollar,'?') : strchrlimit(dollar,tail,'?');
      if (qualifier) {
	*qualifier++='\0'; //!< separate name and qualifier
	char tailchar;
	if (tail) {
	  tailchar=*tail;
	  *tail='\0';
	}
	haspositive=token_search(qualifier,':',true);
	if (tail)
	  *tail=tailchar;
	if (haspositive) {
	  *haspositive++='\0';
	  std::swap(qualifier,haspositive);
	}
      }
    }

    if (flags & SUB_NUMERIC) {      
      const char* varval=NULL;
      if (n>=0) {
	if (curframep->isargdefined(n))
	  varval=curframep->arg(n);	  
      }
      else {
	if (isenvironment(dollar))
	  varval=getstoredenv(dollar);
	else {
	  pushback(curptr,endptr,'$');
	  if (hasparen)
	    *curptr++='(';
	  
	  cptr=origcptr; //not found, continue copy
	  continue;
	}
      }
      cptr=tail+skip;
      if (!varval) { //not defined?
	if (qualifier) {
	  if (n>=0) {
	    assert(tail>=qualifier);
	    *tail='\0';
	  }
	  //	  *parend='\0';
	  int locflags=flags;
	  if (curptr!=buffer)
	    locflags &= ~SUB_ABORTHASH;
	  curptr=substitute_string(curptr,endptr-curptr,qualifier,locflags); //use alternate string (expand recursively)
	  continue;
	}
	//	cptr=tail;
	if ((*buffer=='#') && (flags & SUB_ABORTHASH)) {
	  pushback(curptr,endptr,'$');
	  if (hasparen)
	    pushback(curptr,endptr,'(');
	  if (isstar)
	    pushback(curptr,endptr,'*');
	  pushback(curptr,endptr,dollar);
	  if (hasparen)
	    pushback(curptr,endptr,')');
	  continue;
	}
	parser_printcontext() << "undefined variable: $";
	if (n>=0)
	  std::cerr << n;
	else
	  std::cerr << dollar;
	std::cerr << '\n';
	error_abort();
      }
      if (haspositive) {
	int locflags=flags;
	if (curptr!=buffer)
	  locflags &= ~SUB_ABORTHASH;
	curptr=substitute_string(curptr,endptr-curptr,haspositive,locflags); //use alternate string (expand recursively)
	continue;
      }
      pushback(curptr,endptr,varval);
      if (isstar) { //also need to push additional arguments
	n++;
	while (n<=curframep->nargs()) {
	  pushback(curptr,endptr,' ');
	  pushback(curptr,endptr,curframep->arg(n++));
	}
      }
      // cptr=tail;
      continue;
    }

    bool isconst=true;
    const systemvarmap_type::const_iterator curp(systemvarmap.find(dollar));
    if (curp==systemvarmap.end()) {
      UserVariable* curv=findvariable(dollar);
      isconst=isvarconstant(curv);
      const BaseList<double>& value(curv->get_list());
      if (value.size()==1) {
	snprintf_prec(tmp,sizeof(tmp),value.front());
	pushback(curptr,endptr,tmp);
      }
      else {
	pushback(curptr,endptr,"[");
	const char* startlist=curptr;
	for (size_t i=0;i<value.size();i++) {
	  if (curptr-startlist>substitute_maxchars) {
	    pushback(curptr,endptr,",<list truncated>");
	    break;
	  }
	  if (i)
	    pushback(curptr,endptr,",");
	  snprintf_prec(tmp,sizeof(tmp),value(i));
	  pushback(curptr,endptr,tmp);
	}
	pushback(curptr,endptr,"]");
      }
    }
    else {
      isconst=(isvarconstant(curp->second) || isallowed_nonconst_var(curp->second)); //!< make exception for evaluation index which is uniquely non-const overall, but still const outside execution loop
      pushback(curptr,endptr,curp->second->format());
    }
    if (!isconst && (flags & SUB_NONCONSTWARN))
      nonconstsubs_warning.raise(origcptr);

    cptr=tail;
  }
  *curptr='\0';// ensure terminated
  return curptr;
}

char* parse_string(int flags)
{
  char* cptr=get_token(flags);
  if (!cptr || !(*cptr)) {
    if (flags & F_ALLOWMISSING)
      return NULL;
    error_abort("Null string argument");
  }
  if (strchr(cptr,'$')) {
    char buffer[MAXLINE];
    substitute_string(buffer,sizeof(buffer),cptr);
    return strdup(buffer); //memory leak, but not cumulative
  }
  return cptr;
}

bool parser_isnormal() 
{
  const char* ptr=get_curline();
  return (ptr && *ptr) ? ((*ptr!='-') || isspace(ptr[1]) || !ptr[1]) : false;
}

bool parser_isalpha()
{
  const char* ptr=get_curline();
  return (ptr && *ptr) ? isalpha(*ptr) : false;
}
  
size_t parse_flag(const flagsmap_type& map, const char* cptr)
{
  const flagsmap_type::const_iterator curp(map.find(cptr));
  if (curp==map.end()) {
    parser_printcontext() << "Allowed flags are ";
    dumpflags(std::cerr,map) << '\n';
    error_abort();
  }    
  return curp->second;
}

size_t parse_flags(const flagsmap_type& map, size_t defflags, bool oneonly)
{
  size_t flags=0;
  for (;;) {
    char* cptr= parser_isflag() ? get_token(F_ALLOWMISSING) : NULL;
    if (cptr==NULL)
      return flags ? flags : defflags;
    if (flags & oneonly) {
      parser_printcontext() << "Only one flag allowed out of ";
      dumpflags(std::cerr,map) << '\n';
      error_abort();
    }
    const size_t add=parse_flag(map,++cptr);
    if (flags & add)
      throw InternalError("parse_flags: overlapping flags (or oneonly not set)");
    flags|=add;
  }
}

std::ostream& dumpflags(std::ostream& ostr, const flagsmap_type& map, size_t flags)
{
  flagsmap_type::const_iterator start=map.begin();
  const flagsmap_type::const_iterator end=map.end();

  while (start!=end) {
    if (flags & (start->second)) {
      //      const std::string& fname=(start->first).name;
      ostr << '-' << (start->first) << ' ';
    }
    ++start;
  }
  return ostr;
}

void parser_checkfinished()
{
  if (are_left())
    error_abort("too many arguments");
}

template<> size_t parse_raw(char* cptr, Variable* vardefp, int flags)
{
  const int val=parse_raw<int>(cptr,vardefp,flags);
  if (val<0) {
    parser_printcontext() << cptr << " evaluated to negative integer\n";
    error_abort();
  }
  return val;
}

int parse_unsigned(Variable* vardefp, int flags)
{
  char* cptr=get_token(flags);
  if (!cptr) {
    if (flags & F_ALLOWMISSING)
      return -1;
    error_abort("Missing integer argument");
  }
  return parse_raw<size_t>(cptr,vardefp,flags);
}

template<> int parse_raw(char* cptr, Variable* vardefp, int flags)
{
  return round_int(parse_raw<double>(cptr,vardefp,flags));
}

int parse_int(Variable* vardefp, int flags)
{
  return parse_raw<int>(get_token(),vardefp,flags);
}

LIST<int> parse_intarray(int flags)
{
  static LIST<double> tmparray;
  parse_array(tmparray,get_token(),flags);
  LIST<int> destarray(tmparray.size(),mxflag::temporary);
  for (size_t i=tmparray.size();i--;)
    destarray(i)=round_int(tmparray(i));
  return destarray;
}
  
LIST<size_t> parse_unsignedintarray(int offset, int flags)
{
  static LIST<double> tmparray;
  parse_array(tmparray,get_token(),flags);
  LIST<size_t> destarray(tmparray.size(),mxflag::temporary);
  for (size_t i=tmparray.size();i--;) {
    const int val=round_int(tmparray(i)-offset);
    if (val<0) {
      parser_printcontext() << "list contains elements less than " << offset << '\n';
      error_abort();
    }
    destarray(i)=val;
  }
  return destarray;
}

Variable::Variable(subsid_t subsidv, Setable* ptrv, VariableBase* valuepv)
  : subsid(subsidv), valuep(valuepv)
{
  if (ptrv)
    setptr(ptrv);
  else
    ptr=NULL;
}

const char* Variable::name() const
{
  if (namestr.empty()) {
    validate();
    std::ostringstream str(std::ostringstream::out);
    ptr->printvariablename(str,subsid);
    namestr=str.str();
    if (namestr.empty())
      throw InternalError("Variable name cannot be empty");
  }
  return namestr.c_str();
}

void Variable::setptr(Setable* ptrv)
{
  if (!ptrv)
    throw InvalidParameter("Variable: pointer not set");
  ptr=ptrv;
  if (valuep) {
    if ((verbose & VER_PARSE) && (verbose_level>1)) {
      std::cout << "Creating variable map entry: " << key() << "\n";
      std::cout << (*this) << '\n';
    }
    varmap[key()]=valuep;
  }
}

 void Variable::validate() const
 {
   if (!ptr || !valuep)
     throw InternalError("Variable: pointers not set");
 }

varkey_t Variable::key() const
{ 
  if (!ptr)
    throw InternalError("Variable::key: Setable pointer not set");
  return varkey_t(ptr,subsid);
}

double UserVariable::get() const
{ 
  const BaseList<double>& val(value().value());
  if (val.size()!=1)
    throw Failed("Array cannot be used in this context");
  return val.front();
}

void UserVariable::set(double fval, subsid_t subsid)
{
  assert(subsid==S_ARG1);
  value().set(fval);
}

void UserVariable::set(const BaseList<double>& fvals, subsid_t subsid)
{
  assert(subsid==S_ARG1);
  //   BaseList<double> dest(valuep->value());
  if (fvals.vector()!=value().value().vector()) //!< trap assign to self
    value().set(fvals);
}

// void UserVariable::print(std::ostream& ostr, bool full) const
// {
//   print(ostr,full ? S_NONE : S_ARG1);
// }

void UserVariable::print(std::ostream& ostr) const
{
  ostr << name() << " = ";
  value().print(ostr,true);
  if (isconstant() || (!isused())) {
    ostr << " (";
    if (isconstant())
      ostr << "const" << (isused() ? ")" : ", unused)");
    else
      ostr << "unused)";
  }
}

void UserVariable::printvariablename(std::ostream& ostr, subsid_t) const
{
  ostr << name();
}

template<class T> inline void assign_(BlockedMatrix<T>& to, const BlockedMatrix<T>& from)
{
  //catch assignments that would change matrix size - this must be wrong!
  if ((!!to) && !arematching(to,from))
    throw Failed("Can't assign between matrices of differing size");
  to=from;
}

UserVariable* isvardefined(const char* varchar)
{
  //  const bool islocal=(*varchar=='_');
  //  const uservarmap_type& usemap = islocal ? curframep->localvarmap : vardefmap;
  if (systemvarmap.find(varchar)!=systemvarmap.end())
    error_abort("Can't redefine system variable");
  const uservarmap_type& usemap(vardefmap);
  const uservarmap_type::const_iterator iter=usemap.find(varchar);
  return (iter==usemap.end()) ? NULL : iter->second.get();
}

UserVariable* makevariable(const char* varchar, VariableBase& var)//, bool isconst)
{
  UserVariable* newvarp = new UserVariable(varchar,var);//,isconst);      
  vardefmap[varchar].reset(newvarp);
  return newvarp;
}

UserVariable* UserVariable::create(const char* varchar, double val)
{
  VariableBase* varp=new VariableBase(val);
  return create(varchar,*varp);//,true);
}

const char* parse_variable_name()
{
  return clean_variable_name(parse_string());
}

ContextWarning<> UserVariable::replacingusedwithvar_warning("Replacing already used constant variable with non-const version - this is valid, but not recommended: ",&NMRsim_repeat_warning);
ContextWarning<> UserVariable::replacingunused_warning("Replacing unused (redundant?) definition: ",&NMRsim_once_warning);

UserVariable* UserVariable::create(const char* varchar, VariableBase& var)//, bool isconst)
{
  //  const bool islocal=(*varchar=='_');
  //  uservarmap_type& usemap = islocal ? curframep->localvarmap : vardefmap;
  //const uservarmap_type::iterator iter=usemap.find(varchar);
  //  const bool found=(iter!=usemap.end());
  UserVariable* newvarp=isvardefined(varchar);
  if (!newvarp)
    return makevariable(varchar,var);//,isconst);  
  if (newvarp->isused()) {
    if (newvarp->isconstant()) {
      if (!(var.isconst()))
	replacingusedwithvar_warning.raise(varchar);
    }
    else {
      parser_printcontext() << "can't redefine variable that is in use: " << varchar << '\n';
      error_abort();
    }
  }
  else
    replacingunused_warning.raise(varchar);
    
  newvarp->reset(var);//,isconst);
  return newvarp;
}

size_t parse_counting(const char* str)
{
  char *end;
  size_t n=strtoul(str,&end,10);
  if ((errno==ERANGE) || (str==end))
    n=0;
  return n;
}

size_t maxslot(const char* str)
{
  size_t maxn=0;
  for (;;) {
    str=strchr(str,'#');
    if (str==NULL)
      return maxn;
    char* end;
    str++;
    const long n=strtol(str,&end,10);
    if (end!=str) {//!< ignore invalid - if real problem will be caught when expression is parsed
      if (n<1)
	error_abort("Arguments (#<n>) are numbered from 1");
      if (n>maxn)
	maxn=n;
      str=end;
    }
  }
}

Expression* parse_expression(const char* cptr)
{
  expr_grammar& expr_parser(get_expr_parser());
  expr_parser.reset(F_ALLOWSLOT);
  const parse_info<> info(parse(cptr,
			  expr_parser.use_parser<expr_grammar::expression_def>(),
			  space_p));
  if (!info.full) {
    parser_printcontext() << "expression parsing failed at: \"" << info.stop << "\"\n";
    error_abort();
  }
  Expression& expr(expr_parser.expression());
  if (expr.empty())
    throw InternalError("parse_expression");
  Expression* newexprp=new Expression;
  newexprp->swap(expr);
  return newexprp;
}

namespace {
  //! check whether function (or chained) can be redefined
  void checkredefine(const function_spec& spec)   
  {
    Function_Factory_t& factory(get_Function_Factory());    
    std::string curname(spec.first);

    for (;;curname+=NMRSIM_OLD_FUNCTION) {
      const function_spec curspec(curname.c_str(),spec.second);
      Function_Factory_t::iterator iter(factory.find(curspec));
      if (iter==factory.end()) 
	return; //!< not found
    
      const function_def_t& fdef(iter->second);
      if (fdef.used) {
	parser_printcontext() << "can't redefine function previously used (as non const): " << curname << '\n';
	error_abort();
      }    
      function_redefinition_warning.raise(curname.c_str());
    }
  }

  //! do redefinition
  void doredefine(const function_spec& spec, ExpressionUserFunction* funcp)
  {
    function_def_t fdef(funcp);
    Function_Factory_t& factory(get_Function_Factory());    
    std::string curname(spec.first);

    for (;;curname+=NMRSIM_OLD_FUNCTION) {
      const function_spec curspec(curname.c_str(),spec.second);
      Function_Factory_t::iterator iter(factory.find(curspec));
      if (iter==factory.end()) {
	std::string* cnamep=new std::string(curname);	
	factory[function_spec(cnamep->c_str(),spec.second)]=fdef;
	return;
      }
      ::std::swap(fdef,iter->second);
    }
  }
}

ContextWarning<> slotmismatch_warning("mismatch between number of argument types specified and number of arguments used in function: ",&NMRsim_once_warning);

void parse_function()
{
  static LIST<const Expression*> expr_pointer_stack;

  const std::string* strnamep=new std::string(parse_string()); //!< need to create new string rather than pointer
  const char* funcname=strnamep->c_str();
  const char* exprval=parse_string();

  argdef_t argtype=0;
  size_t argcount=0;
  if (are_left()) {
    const char* argdef=parse_string();
    argcount=strlen(argdef);
    for (size_t i=argcount;i--;) {
      argtype<<=1;
      switch (argdef[i]) {
      case 'S':
	break;
      case 'L':
	argtype|=1;
	break;
      default:
	parser_printcontext() << "invalid argument type specification (" << argdef << ").  Each argument should be one of S (scalar) or L (list)\n";
	error_abort();
      }
    }
    if (slotmismatch_warning.enabled()) {
      if (maxslot(exprval)!=argcount)
	slotmismatch_warning.raise(funcname,true);
    }
  }
  else {
    argcount=maxslot(exprval);
//     if (argcount==0)
//       nofunctionarguments_warning.raise(funcname,true);
  }
  const function_spec spec(funcname,argcount);
  checkredefine(spec);

  expr_pointer_stack.push_back((const Expression*)NULL); //!< create space for new indirect pointer
  ExpressionUserFunction* funcp=new ExpressionUserFunction(spec,argtype,expr_pointer_stack.back());
  doredefine(spec,funcp);
  const Expression& expr(*parse_expression(exprval));
//   if (expr.isconstant())
//     constfunction_warning.raise(funcname,true);
  funcp->setexpression(expr);
  if (verbose & VER_PARSE) {
    std::cout << "Creating function " << funcname << " with arguments ";
    for (size_t i=0;i<funcp->nargs();i++)
      std::cout << (funcp->islist(i) ? 'L' : 'S');
    std::cout << ": ";
    expr.print();
    std::cout << '\n';
  }
}

void parse_variable()
{
  LIST<const char*> varnames;
  char* varcharall=parse_string();
  Splitter splitter(varcharall,",;",true);
  char* varchar;
  char tupleseparator;
  while ((varchar=splitter.next(tupleseparator))) {
//     const bool islocal=(*varchar=='_');
//     if (islocal) {
//       if ((curframep==topframep) && !nochecks)
// 	parser_printcontext() << "warning: local variable declared at top level and so effectively global (block level scoping intended?)\n-nochecks disables this warning\n";
//     }
//     else {
    if (!isalpha(*varchar))
      error_abort("Variable name must start with letter");
    const systemvarmap_type::iterator iter(systemvarmap.find(varchar));
    if (iter!=systemvarmap.end()) {
//       const SystemVariableBase& var(*(iter->second));
//       std::cerr << "Hash: " << iter->first << "  ";
//       var.print(std::cerr); std::cerr << '\n';
//       std::cerr << "Hash of " << varchar << ": " << systemvarmap.hash(varchar) << std::endl;
      error_abort("Cannot redefine existing (system) variable");
    }
      
    if (isenvironment(varchar))
      error_abort("Variable name must contain lower case letters (to prevent clash with environment variables)");
    
    for (const char* p=varchar+1;*p;p++) {
      if (!isalpha(*p) && !isdigit(*p) && (*p!='_'))
	error_abort("Variable name can only contain letters, digits or _");
    }  
    if ((strcmp(varchar,"x")==0) || (strcmp(varchar,"y")==0))
      xy_variable_warning.raise();

    varnames.push_back(varchar);
  }
  if (varnames.empty())
    error_abort("no variables to define!");
        
  char* cptr=get_token();
  if (checkbracket(cptr,'\'')) {
    if (varnames.size()!=1)
      error_abort("strings can't be used in tuple assignment");    
    const char* varchar(varnames.front());
    if (*varchar=='_')
      error_abort("string variables cannot be local");
    systemvarmap[varchar]=new SystemVariable<std::string>(varchar,cptr);
    return;
  }

  Mark markobj;
  const bool istuple=(varnames.size()>1);
  UserVariable* oldvarp=istuple ? NULL : isvardefined(varnames.front());
  //Postpone until UserVariable::create ?
//   if (oldvarp && oldvarp->islinked())
//     error_abort("can't redefine variable that is used");
  Variable cvar(S_ARG1,oldvarp);
  parse_double_raw(&cvar,cptr,F_ALLOWLIST | F_ALWAYSNEW); //!< get value
  const bool isconst=(cvar.subsid==S_NONE); //!< will be flagged as non-const even if only one element of list is non-const
  VariableBase* valuep=cvar.valuep;
  assert(valuep!=NULL);
  LIST<double> tmpvars;
  const size_t nvalues=valuep->value().size();
  if (istuple) {
    if (isconst) {
      if ((tupleseparator!=';') && (nvalues % varnames.size())) {
	parser_printcontext() << "bad tuple assignment: " << (valuep->value().size()) << " values cannot be split evenly between " << varnames.size() << " variables\n";
	error_abort();
      }
//       if (varnames.size()>nvalues) {
// 	parser_printcontext() << "bad tuple assignment: fewer values (" << nvalues << ") than destination variables (" << varnames.size() << '\n';
// 	error_abort();
//       }
      tmpvars.create((nvalues+varnames.size()-1) / varnames.size());
    }
    else {
      if (!(valuep->isexpr()))
	error_abort("Can only use tuple assignment for constants / expressions");
    }
  }
  static ScratchList<size_t,2> extargs(3U);
  static size_t tuplecount=0;
  UserVariable* tuplevarp=NULL;
  if (istuple && !isconst) {//!< create base tuple expression
    tuplecount++;
    char tmpname[16];
    sprintf(tmpname,"<tuple%i>",tuplecount);
    tuplevarp=UserVariable::create(tmpname,*valuep);
    tuplevarp->flagused(); //!< always used
    markobj.flush(tuplevarp);
  }
  size_t valuesleft=nvalues;
  const bool commasep=(tupleseparator==',');
  for (size_t j=0;j<varnames.size();j++) {
    VariableBase* usevarp=valuep;
    const char* varchar(varnames(j));
    if (istuple) {
      if (isconst) {
	const BaseList<double> source(valuep->value());
	const size_t n=tmpvars.size();
	if ((n==1) && commasep)
	  usevarp=new VariableBase(source(j));
	else {
	  if (n) {
	    if (commasep)
	      tmpvars=source(libcmatrix::range(n*j,n*j+n-1));
	    else {
	      const size_t items=(valuesleft<n) ? valuesleft : n;
	      valuesleft-=items;
	      if (items) {
		const slice sl(j,items,varnames.size());
		tmpvars=source(sl);
	      }
	      else
		tmpvars.clear();
	    }
	  }
	  usevarp=new VariableBase(tmpvars);
	}
      }
      else { //!< can assume that valuep refers to an expression	
	Expression* newexpr = new Expression;
	extargs(0U)=newexpr->push(new ExpressionVariable(*tuplevarp));
// 	extargs(1U)=newexpr->push(new ExpressionConstant(j+1));
// 	newexpr->push(new FunctionExtract(extargs));
	extargs(1U)=newexpr->push(new ExpressionConstant(varnames.size()));
	extargs(2U)=newexpr->push(new ExpressionConstant(j+1));
	if (tupleseparator==',')
	  newexpr->push(new FunctionSplitRow(extargs));
	else
	  newexpr->push(new FunctionSplitColumn(extargs));
	usevarp=new VariableBase;
	usevarp->initialise(*newexpr);
	Variable var(S_ARG1,isvardefined(varchar),usevarp);
	register_expression(var);
      }
    }
    markobj.flush(UserVariable::create(varchar,*usevarp));
  }
}

std::ostream& parser_printcontext(std::ostream& ostr, bool printtoken)
{
  const bool havetoken=((lasttoken!=NULL) && printtoken);
  if (curframep)
    ostr << "At " << curframep->state();
  if (havetoken)
    ostr << (curframep ? ", parsing " : "While parsing ") << lasttoken;
  if (havetoken || curframep)
    ostr << ": ";
  return ostr;
}

void snprintf_prec(char* buf, size_t s, double val)
{
  static char formatstr[10]="";
  static bool hasprec=false;
  if (*formatstr=='\0') {
    const char* nmrsimformat=getenv("NMRSIM_FORMAT");
    const bool useexplicit=(nmrsimformat!=NULL) && (*nmrsimformat!='\0');
    if (useexplicit) {
      formatstr[0]='%';
      if (strchr(nmrsimformat,'%'))
	error_abort("NMRSIM_FORMAT should not contain %.  Examples of valid usage: g, f, .*g etc.");
      strncat(formatstr+1,nmrsimformat,sizeof(formatstr)-2);
    }
    else
      strcpy(formatstr,"%.*g");
    if ((verbose & VER_GEN) && (verbose_level>1)) {
      std::cout << "Setting numeric formatting to " << formatstr << " (";
      if (useexplicit)
	std::cout << "from NMRSIM_FORMAT)\n";
      else
	std::cout << "default)\n";
    }
    hasprec=(strchr(formatstr,'*')!=NULL);
  }
  if (hasprec) {
    const ostream_controller& ctrl(cmatrix_ostream_controller());
    const int precision=(ctrl.matrixprecision<0) ? std::cout.precision() : ctrl.matrixprecision;
    snprintf(buf,s,formatstr,precision,val);
  }
  else
    snprintf(buf,s,formatstr,val);
}

void parse_ignored()
{
  set_curline(NULL);
  ignoreddirective_warning.raise();
}


double Variable::get() const
{
  const BaseList<double>& value(valuep->value());
  if (value.size()!=1)
    throw Failed("Can't return list expression as single value");
  return value.front();
}

const char* Variable::format() const
{
  const BaseList<double>& value(valuep->value());
  if (value.size()>1)
    strcpy(buf,"<list>");
  else
    snprintf_prec(buf,sizeof(buf),value.front());
  return buf;
}

void Variable::printname(std::ostream& ostr) const
{
  if (!ptr)
    throw InternalError("Variable: ptr unset");  
  ptr->printvariablename(ostr,subsid);
}

void Variable::print(std::ostream& ostr, bool full) const
{
  printname(ostr);
  ostr << ": ";
  valuep->print(ostr,full);
}

// namespace {
//   template<class T> T* unwrap(T* v) { return v; }
//   template<class T> T* unwrap(const smartptr<T,false>& v) { return v.get(); }
// }

template<class T> void dumpvariables(const T& varmap, const char* type)
{
  std::cerr << "Currently defined " << type << " variables:";
  typename T::const_iterator start(varmap.begin());
  const typename T::const_iterator end(varmap.end());
  if (start==end)
    std::cerr << " <none>";
  else {
    while (start!=end) {
      std::cerr << ' ' << start->second->name();
      ++start;
    }
  }
  std::cerr << std::endl;
}

UserVariable* findvariable(const char* name, bool flagused)
{
  const uservarmap_type::const_iterator curp(vardefmap.find(name));
  const uservarmap_type::const_iterator end(vardefmap.end());
  if (curp==end) {
    if (isenvironment(name)) {
      parser_printcontext() << "unsubstituted environment variable: " << name << "\nUse $(" << name << ") to separate substitution variable from following text\n";
      error_abort();
    }
    parser_printcontext() << "variable " << name << " not defined\n";
    if (noexecute)
      std::cerr << "(NB. user-defined variables are not created with -noexecute)\n";
    else {
      dumpvariables(vardefmap,"user-defined");
      dumpvariables(systemvarmap,"system");
    }
    error_abort();
  }
  UserVariable* retval=(curp->second).get();
  if (flagused)
    retval->flagused();
  return retval;
}

UserVariable* findnewvariable(const char* name)
{
  const uservarmap_type::const_iterator curp(vardefmap.find(name));
  const uservarmap_type::const_iterator end(vardefmap.end());
  if (curp!=end) {
    parser_printcontext() << "variable " << name << " already in use - must use a fresh variable name in this context\n";
    error_abort();
  }
  if (isenvironment(name))
    error_abort("variable name must contain lowercase characters to distinguish from environment variable");
  VariableBase* varp=new VariableBase();
  return makevariable(name,*varp);//,false);
}

 void dimension_set::set(size_t dim, size_t n)
{
  if (dim>MAX_DIMENSIONS) {
    parser_printcontext() << "dimension (" << dim << ") out of range.  Maximum is " << MAX_DIMENSIONS << '\n';
    error_abort();
  }
  dims[dim].set(n,block_change);
}

ContextWarning<> rowsskip_warning("total number of rows is not multiple of 'skip factor'",&NMRsim_repeat_warning);

void dimension_set::set(size_t dim, size_t n, size_t skip)
{
  if (dim>MAX_DIMENSIONS) {
    parser_printcontext() << "dimension (" << dim << ") out of range.  Maximum is " << MAX_DIMENSIONS << '\n';
    error_abort();
  }
  if (dim==0)
    throw InvalidParameter("Can't set indirect dimension 0");
  if (block_change)
    error_abort("cannot set indirect dimension at this point (already fixed)");
  if (n<1)
    error_abort("dimension cannot be <1");
  if (skip<1)
    error_abort("row step cannot be <1");
    //    skips(dim-1)=ni_skip;
  if (n % skip)
    rowsskip_warning.raise();
  dims[dim].set(n,skip);
}

void dimension_set::dimension::set(size_t n, size_t skipv)
 {
   if (n==0)
     throw InvalidParameter("Dimension cannot be zero");
   if (hard)
     error_abort("cannot change size of indirect dimension");
   hard=n;
   skip=skipv;
 }

void dimension_set::dimension::set(size_t n, bool block)
 {
   if (n==0)
     throw InvalidParameter("Dimension cannot be zero");
   
   if ((n==dim) || (n==1))
     return; //!< no change

   if (dim && ((n<dim) ? (dim % n) : (n % dim)))
     incommensurate_warning2.raise(state);
   
   if (n>=dim) {
     if (block) {
       parser_printcontext() << "cannot change dimension (from " << dim << " to " << n << ") at this point (already fixed)\n";
       error_abort();
     }
     dim=n;
     state=curframep->state();
   }
 }

void dimension_set::dimension::error(size_t dim, const char* err)
{
  parser_printcontext() << "dimension " << dim << ": " << err << '\n';
  error_abort();
}

size_t dimension_set::dimension::get(size_t& array_skip, size_t N) const
{
  array_skip=1;
  if (hard) {
    if (dim) { //!< also soft set?
      if ((hard+skip-1)/skip==dim) //!< need skip to make things match?
	array_skip=skip;
      else {
	if (dim!=hard)
	  error(N,"array and nD size are incommensurate");
      }
    }
    else //!< by default follow any "quadrature" setting
      array_skip=skip;
    return hard;
  }
  if (!dim)
    error(N,"not set");
  return dim;
}

size_t dimension_set::get(LIST<size_t>& dest, LIST<size_t>& array_skip, size_t& totdim) const
{  
  size_t maxdim=MAX_DIMENSIONS;
  while (maxdim && !dims[maxdim].isset())
    maxdim--;
  totdim=1;
  dest.create(maxdim);
  array_skip.create(maxdim);
  size_t nhard=0;
  for (size_t dim=maxdim;dim;dim--) { //check for "holes"
    size_t& curskip(array_skip(dim-1));
    size_t n=dims[dim].get(curskip,dim);
    if (dims[dim].ishard())
      nhard++;
    totdim*=n;
    dest(dim-1)=n;
  }
  const size_t dim0=dims[0].dim;
  if (dim0) {
    if (maxdim) {
      bool incom;
      if (dim0>totdim) {
	if (active2D) {
	  std::cerr << "Data set size smaller than number of rows set at " << dims[0].state << '\n';
	  error_abort();
	}
	incom=dim0 % totdim;
	totdim=dim0;
      }
      else
	incom=totdim % dim0;
      
      if (incom)
	incommensurate_warning.raise(dims[0].state);
    }
    else //not nD, but arrayed
      totdim=dim0;
  }
  return nhard;
}

void stripleaf(char* fname,const char* strip)
{
  char* poss_start=fname+strlen(fname)-strlen(strip);
  if ( (poss_start>=fname) && (strcmp(poss_start,strip)==0))
    *poss_start='\0';
}

bool read_block(const char* name,command_Factory_t& fac, bool isopt, bool ignoreunk)
{
  //  static command_Factory_t& Global_Factory(get_Global_Factory());
  parser_newblock(name,isopt);
  bool foundblock=false;
  while (parser_getline()) {
    foundblock=true;
    const char* curcom(parse_string());
    command_Factory_t::iterator curp=fac.find(curcom);
    if (curp==fac.end()) {
      if (ignoreunk) {
	if (verbose & VER_PARSE)
	  parser_printcontext() << "ignoring " << curcom << '\n';
	continue;
      }
      else {	
	if (*curcom=='#')
	  error_abort("directive cannot begin with # - misplaced comment? (# must be first character of line)");
	handle_unrecognised(curcom);
      }
    }
    try {
      (curp->second)();
      parser_checkfinished();
    } catch (MatrixException& exc) {
      std::cerr << exc;
      std::cerr.flush();
      error_abort();
    }
    checkdirty();
  }  
  return foundblock;
}

template<class T> void parse_system_variable(SystemVariable<T*>& gvar,int flags)
{
  Mark markobj;
  Variable cvar(S_ARG1,&gvar);
  gvar=parse_raw<T>(get_token(),&cvar,flags);
  const bool isconst=(cvar.subsid==S_NONE);    
  if (isconst)
    gvar.isconstant(true); //! ensure const status
  else {
    gvar.reset(cvar.variable());
    markobj.flush(&gvar);
  }
}

template void parse_system_variable(SystemVariable<double*>&, int);
template void parse_system_variable(SystemVariable<size_t*>&, int);
template void parse_system_variable(SystemVariable<int*>&, int);

//! round \c double to nearest integer
/** exit with ::ERR_FAILED if not within ::NMRSIM_ROUNDTOL */
int round_int(double f)
{
  const int i=rawround_int(f);
  if (fabs(f-i)>NMRSIM_ROUNDTOL) {
    parser_printcontext() << "quantity (" << f << ") does not round to integer\n";
    error_abort();
  }
  return i;  
}

//NB silently fails if str is not a valid key!
bool havekey(const command_Factory_t& fac, const std::string& str)
{
  const command_Factory_t::const_iterator pos(fac.find(str));
  if (pos==fac.end()) {
    std::cerr << "havekey: key doesn't exist: " << str << '\n';
    error_abort();
  }
  return pos->second.done; 
}

void verify_arraysize(size_t n, size_t dim)
{
  if (n!=1)
    array_dims.set(dim,n);
}

void ExpressionVariableFunction::create(BaseList<std::string>& args)
{
  if (args.size()!=1)
    throw InternalError("Functions operating on variables only take one argument");
  varp=findvariable(args.front().c_str());
  const VariableBase rawvar(varp->value());
  if (rawvar.isexpr())
    appliedtoexpr_warning.raise();
}
  
ContextWarning<> ExpressionVariableFunction::outofmainloop_warning("variable contents not well defined in this context: ",&NMRsim_once_warning);
ContextWarning<> ExpressionVariableFunction::outofmainloop_expression_warning("expression may not be well defined in this context: ",&NMRsim_once_warning);

void FunctionValueof::operator()(LIST<double>& dest) const
{
  assert(varp!=NULL);
  dest=varp->get_list();
}

bool FunctionValueof::rawisconstant() const
{
  assert(varp!=NULL);
  return varp->isconstant();
}

void FunctionValuesof::operator()(LIST<double>& dest) const
{
  assert(varp!=NULL);
  const VariableBase val(varp->value());
  if (val.isexpr())
    throw Failed("Valuesof applied to expression");
  if (val.isarray()) {
    if (val.isexprarray())
      throw Failed("Valuesof applied to non-const array");
    dest=val.get_array().row();
  }
  else
    dest=val.value();
}

//ContextWarning<> ExpressionVariableFunction::nonarray_warning("Valuesof applied to non-array variable: ",&NMRsim_once_warning);

bool FunctionValuesof::rawisconstant() const
{
//   assert(varp!=NULL);
//   if (!(varp->value().isarray()) || varp->value().isexprarray())
//     nonarray_warning.raise(varp->name());
  return true;
}
 
void FunctionErrorof::operator()(LIST<double>& dest) const
{
  assert(varp!=NULL);
  double err=varp->get_error();
  if (err<0.0)
    err=0.0;
  dest.create(1U,err);
}

bool FunctionErrorof::rawisconstant() const
{
  assert(varp!=NULL);
  return (varp->value().issimple());
} 
bool FunctionErrorsof::rawisconstant() const { return true; }

void FunctionErrorsof::operator()(LIST<double>& dest) const
{
  assert(varp!=NULL);
  dest=varp->get_errors();
  for (size_t i=dest.size();i--;) {
    if (dest(i)<0.0)
      dest(i)=0.0;
  }
}

void ExpressionVariableFunction::rawprint(std::ostream& ostr) const
{
  assert(varp);
  ostr << name_ << "(\'" << varp->name() << "\')";
}

bool VariableBase::validate_context(const char* name) const
{
  if (isconst() || within_evaluation())
    return true;
  const bool arrayvalid=(evaluation_state==CONTEXT_POSTCALC) || (evaluation_state==CONTEXT_FINALISE);
  if (arrayvalid) { //!< simple arrays are OK within processing blocks
    if (isexpr()) {
      if (sum_n0<2) //!< no summation arrays - always OK
	return true;
    }
    else {
      if (isarray() && (array.size()==1))
	return true;
    }
  }
  if (isexpr())
    ExpressionVariableFunction::outofmainloop_expression_warning.raise(name,true); //!< slightly weak - have to give warning in case expression depends on summation variable
  else {
    if (isarray())
      ExpressionVariableFunction::outofmainloop_warning.raise(name,true);    
  }
  return false;
}

ContextWarning<> nonenvironment_warning("setenv variable name will not be recognised in pNMRsim as environment variable (contains lower case characters)",&NMRsim_repeat_warning);

void parse_setenv()
{
  //char totalcom[4096];
  char* cptr=get_curline();
  chewwhite(cptr);
  char* tail=cptr;
  while (*tail && !isspace(*tail))
    tail++;
  char* storeptr=tail;
  const char storechar=*storeptr;
  if (storechar)
    *tail++='\0';
  if (strchr(cptr,'='))
    error_abort("environment variable name can not contain =");
  chewwhite(tail);
  //if (*tail=='\0')
  //  error_abort("syntax: setenv <variable name> <definition>");
  
  if (!isenvironment(cptr))
    nonenvironment_warning.raise();
  
  //  snprintf(totalcom,sizeof(totalcom),"%s=%s",cptr,tail);  
  if (setenv(cptr,tail,1)) { //!< no need to create copy for setenv
    perror("setenv");
    error_abort();
  }
  
  *storeptr=storechar; //!< restore so original command line preserved
}

void scan_block(const char* name, bool isopt, bool allowmult)
 {
   for (;;) {
     parser_newblock(name,isopt);
     bool done=false;
     while (parser_getline(false)) {
       if (!parser_inblock())
	 std::cout << get_curline() << '\n';
       else {
	 if (!done) {
	   std::cout << name << " {\n";
	   done=true;
	 }
	 std::cout << OUTPUT_SPACER << get_curline() << '\n';
       }
     }     
     if (!done)
       return;
     std::cout << "}\n";
     if (!allowmult)
       return;
   }
 }  
  
