#ifndef Parser_h_
#define Parser_h_

/*! \file
 \brief  Declarations for input file parsing
*/

#include "NMRsim.h"

#ifndef PATH_SEPARATOR
//! separator between path elements 
#define PATH_SEPARATOR ':'
#endif
#ifndef DIR_SEPARATOR
//! separator charactor within filename
#define DIR_SEPARATOR '/'
#endif

#define MAXLINE 4096 //!< max length of input lines
#define MAXPATH 4096 //!< max length of path name

void parser_init(const char*, int argc, char* argv[], int flags =0); //!< initialise parser
bool parser_inblock(); //!< \c true if parsing lines inside a block (as opposed to global)
const char* parser_getfname(); //!< get input file
void parser_newblock(const char* blockname_, bool, const char* keynames_[] =NULL, size_t nkeys_ =0); //!< open new {} block
void parser_getinteractive(const char* intro, std::istream& =std::cin); //!< get parse line from input stream with prompt \a intro 
bool parser_foundlast(); //!< return \c true if last block was found
bool parser_isopen(); //!< return \c true if input file open
const char* clean_variable_name(char*); //!< strip any leading $, warn missing literal quotes
bool parser_getline(bool executeglobal =true); //!< get input line (NULL if no more)
const char* parse_variable_name(); //!< get variable name
size_t parse_counting(const char*); //!< parse counting number (from 1) from string (0 if invalid)
void parser_insertline(const char*);
void parser_flush(); //!< finish with parser
void parser_checkfinished(); //!< generate error if arguments are still left
bool parser_isflag(); //!< return \c true if next token is a flag
bool parser_isnormal(); //!< return \c true if next token is a normal (non-flag) argument
bool parser_isalpha(); //!< return \c true if next token starts with alphabet character (i.e. possible variable name)
void parse_ignored(); //!< dummy for ignored instruction
Expression* parse_expression(const char*); //!< create new expression
int parse_index(size_t nspins_cell, size_t nspins_total =0); //!< parse spin number
double parse_double_raw(Variable*, char* cptr, int flags); //!< parse real quantity (raw)
bool read_block(const char* name,command_Factory_t& fac, bool isopt =false, bool ignoreunk =false); //!< parse {} block 

extern ContextWarning<> path_unreadable_warning;
extern ContextWarning<> path_emptyelement_warning;
extern ContextWarning<> sharedtarget_warning; //!< different variables appear to point to same target
extern ContextWarning<> postcontinuation_whitespace_warning; //!< whitespace after \\ continuation
extern ContextWarning<> extradollar_warning; //!< variable *names* shouldn't have leading $
extern ContextWarning<> namequote_warning; //!< variable names are best quoted
extern ContextWarning<> filetail_ignored_warning; //!< rest of input file is being ignored
extern ContextWarning<> function_redefinition_warning; //!< function is (potentially) being redefined
extern ContextWarning<> misplaced_sum_warning; //!< sum array not meaningful in context
extern ContextWarning<> includeonce_withlist_warning; //!< list arguments used with includeonce
extern ContextWarning<> nofunctionarguments_warning; //!< function with no arguments
extern ContextWarning<> xy_variable_warning; //!< warn about using x or y as variable names
extern ContextWarning<> trailingcharacters_warning; //!< trailing characters after } block end
extern ContextWarning<> ignoreddirective_warning; //!< directive is being ignored
extern ContextWarning<> recursion_warning; //!< excessive recursion warning
extern ContextWarning<> zerolengthscalar_warning; //!< zero length argument when scalar expected
extern ContextWarning<> include_nosubstitution_warning; //!< dynamic include but nothing varies

typedef std::map<std::string, smartptr<UserVariable,false> > uservarmap_type;
extern uservarmap_type vardefmap; //!< registry for user variables

typedef MAPTYPE(size_t) flagsmap_type; //!< type for set of flags

void declare_builtin_block(const char*); //!< register standard block

std::ostream& dumpflags(std::ostream&, const flagsmap_type&, size_t =0xFFFFFFFF); //!< stream flags set

extern LIST<productoperator_spec*> opspecstack; //!< stack of pointers to all operator specifications
//extern bool nondiagonal_opspec; //!< \c true if created non-diagonal product operator
void stripleaf(char* a, const char* trail); //!< strip \a trail from end of \a a if present
UserVariable* findnewvariable(const char* name); //!< create new (empty) ::UserVariable (error if already defined)
char* parse_string(int flags =0); //!< parse string
extern bool iscommonPAS; //!< \c true if all interactions share a common PAS
LIST<int> parse_intarray(int =0); //!< parse list of integers
LIST<size_t> parse_unsignedintarray(int offset =0, int  flags =0); //!< parse list of unsigned integers
int parse_int(Variable* =NULL, int =0); //!< parse integer quantity
int parse_unsigned(Variable* =NULL, int =0); //!< parse unsigned integer
double parse_double(Variable* =NULL, int =0); //!< parse real quantity
double parse_shift(Variable*, double gamrat, int flags =0); //!< parse shift quantity
VariableBase* parse_double_variable(Variable&, int =0);
template<class T> T parse_raw(char*, Variable*, int); //!< \internal raw parse
size_t parse_flag(const flagsmap_type&, const char*); //!< parse individual flag
size_t parse_flags(const flagsmap_type&, size_t def =0, bool oneonly =false); //!< parse flags
char* get_curline(); //!< get pointer to current parse line
void set_curline(char *); //!< set current parse point
char* get_token(int flags =0); //!< get next token (and advance)
size_t count_left(bool =true); //!< count remaining tokens
bool are_left(); //!< return \c true if unparsed tokens left
//PhasedSequence* parse_phasedsequence(int flags =0); //!< parse sequence+phase (NB be careful with setable phase issues)
class CycledSequence;
CycledSequence* parse_cycledsequence(int flags =0); //!< parse cycled sequence
CycledSequence* parse_cycledsequence(char*, int flags =0); //!< parse cycled sequence
double get_nmrfreq(size_t qualifier =0); //!< return current (or specified) NMR frequency
void scan_block(const char* name, bool isopt =false, bool allowmult =false);
void dumpenvironment(std::ostream& ostr); 

enum pm_flags { PM_FULL=1 };
const flagsmap_type& putmatrix_flags();

struct Splitter {
  explicit Splitter(char* source, const char* charsv =",",  bool =false);
  
  char* next(char&);
  char* next() { char junk; return next(junk); } //discard
  bool atend() const { return (curptr==NULL || !(*curptr)); }
  
  char* curptr;
  const char* chars;
  char* s;
  char sepchar;
  bool unique;
};

//! mark point for creation of new variables
class Mark {
public:
  Mark();
  ~Mark();
  void flush(Setable*); //!< flush (set pointer for unassigned variables)
  bool needsflush() const; //!< \c true if unassigned variables
  static ContextWarning<> unflushed_warning; //!< warning if object destroyed before flush
private: 
  size_t stackpos; //!< starting pointer into dirty link stack
};

//! strip leading white space
inline void chewwhite(char*& str) {
  while (isspace(*str)) { str++; } 
}

void verify_arraysize(size_t n, size_t dim =0); //!< register {} array size

template<class T> void verify_arraysizes(const ListList<T>& a, size_t sumtag, const BaseList<size_t>& arraytags)
{
  const size_t lists=a.size();
  if (lists>1)
    sum_dims.set(sumtag,lists);
  for (size_t i=lists;i--;) {
    const size_t dim = arraytags.empty() ? 0 : arraytags(i);
    verify_arraysize(a.size(i),dim);
  }
}

//const char* parser_getdirective(char* buf); //!< set parse line to \a buf and extract directive name

//! find key in parse command factory
template<class Factory> typename Factory::mapped_type factory_parse(const char* keyname, const Factory& factory)
{
  const typename Factory::const_iterator curcom=factory.find(keyname);
  if (curcom==factory.end()) {
    handle_unrecognised(keyname);
    error_abort();
  }       
  return curcom->second;
}

//!< call parsing function creating new object
template<class Func,class T> T* factory_call(Func funcp, Type2Type<T>)
{
  T* newac=NULL;
  Mark markobj;
  try {
    newac=(*funcp)();
    if (newac)
      parser_checkfinished();
  } catch (MatrixException& exc) {
    parser_printcontext() << exc << '\n';
    error_abort();
  }
  if (markobj.needsflush()) {
    assert(newac!=NULL);
    newac->isconstant(false); //flag is not const
    markobj.flush(newac);
  }
  return newac;
}

//!< find and call parsing function from key name
template<class Factory, class T> T* factory_parse(const char* keyname, const Factory& factory, Type2Type<T> type)
{
  const typename Factory::mapped_type funcp(factory_parse(keyname,factory));
  return factory_call(funcp,type);
}

//!< read complete block given parsing factory
template<class StackT, class Factory, class T> bool factory_read(StackT& pstack, const char* name, const Factory& factory, Type2Type<T> type)
{
  if (!parser_isopen())
    return false;
  parser_newblock(name,true);
  bool found=false;
  while (parser_getline()) {
    const char* keyname=parse_string();
    T* newproc=factory_parse(keyname,factory,type);
    if (newproc) {
      pstack.push_back(newproc);
      found=true;
    }
  }
  return found;
}

//!< read blocks (which may be repeated) using given parsing factory
template<class StackT, class Factory, class T> size_t multiple_factory_read(LIST<StackT>& mstack, const char* name, const Factory& factory, Type2Type<T> type)
{
  if (!parser_isopen())
    return 0U;
  size_t count=0U;
  for (;;count++) {
    mstack.push_back(); //!< call constructor for set up
    if (!factory_read(mstack.back(),name,factory,type)) {
      mstack.pop_back();
      return count;
    }
    mstack.back().initialise(); //!< finished parsing, do any tidy up
  }
}

#endif
